"""
Class for reading data from Neuralynx files.
This IO supports NCS, NEV, NSE and NTT file formats.


NCS contains the sampled signal for one channel
NEV contains events
NSE contains spikes and waveforms for mono electrodes
NTT contains spikes and waveforms for tetrodes

All Neuralynx files contain a 16 kilobyte text header followed by 0 or more fixed length records.
The format of the header has never been formally specified, however, the Neuralynx programs which
write them have followed a set of conventions which have varied over the years. Additionally,
other programs like Pegasus write files with somewhat varying headers. This variation requires
parsing to determine the exact version and type which is handled within this RawIO by the
NlxHeader class.

Ncs files contain a series of 1044 byte records, each of which contains 512 16 byte samples and
header information which includes a 64 bit timestamp in microseconds, a 16 bit channel number,
the sampling frequency in integral Hz, and the number of the 512 samples which are considered
valid samples (the remaining samples within the record are invalid). The Ncs file header usually
contains a specification of the sampling frequency, which may be rounded to an integral number
of Hz or may be fractional. The actual sampling frequency in terms of the underlying clock is
physically determined by the spacing of the timestamps between records.

These variations of header format and possible differences between the stated sampling frequency
and actual sampling frequency can create apparent time discrepancies in .Ncs files. Additionally,
the Neuralynx recording software can start and stop recording while continuing to write samples
to a single .Ncs file, which creates larger gaps in the time sequences of the samples.

This RawIO attempts to correct for these deviations where possible and present a single section of
contiguous samples with one sampling frequency, t_start, and length for each .Ncs file. These
sections are determined by the NcsSectionsFactory class. In the
event the gaps are larger, this RawIO only provides the samples from the first section as belonging
to one Segment.

If .Ncs files are loaded these determine the Segments of data to be loaded. Events and spiking data
outside of Segments defined by .Ncs files will be ignored. To access all time point data in a
single Segment load a session excluding .Ncs files.

Continuous data streams are ordered by descending sampling rate.

This RawIO presents only a single Block.

Author: Julia Sprenger, Carlos Canova, Samuel Garcia, Peter N. Steinmetz.
"""

from ..baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)
from operator import itemgetter
import numpy as np
import os
import pathlib
import copy
import warnings
from collections import namedtuple, OrderedDict

from neo.rawio.neuralynxrawio.ncssections import NcsSection, NcsSectionsFactory
from neo.rawio.neuralynxrawio.nlxheader import NlxHeader


class NeuralynxRawIO(BaseRawIO):
    """
    Class for reading datasets recorded by Neuralynx.

    This version works with rawmode of one-dir for a single directory of files or one-file for
    a single file.

    Parameters
    ----------
    dirname: str, default: ''
        Name of directory containing all files for a dataset. If provided, filename is
        ignored. But one of either dirname or filename is required.
    include_filenames: str | list | None, default: None
        Name of a single ncs, nse, nev or ntt file or list of such files. Will only include
        file names in the list. This can be plain filenames or fullpath or path relative to
        dirname.
        All files should be in a single path.
    exclude_filenames: str | list | None, default: None
        Name of a single ncs, nse, nev or ntt file or list of such files. Expects plain
        filenames (without directory path).
        None will search for all file types
    keep_original_times: bool, default: False
        If True, keep original start time as in files,
        Otherwise set 0 of time to first time in dataset
    strict_gap_mode: bool, default: True
        Detect gaps using strict mode or not.
          * strict_gap_mode = True then a gap is consider when timstamp difference between two
            consequtive data packet is more than one sample interval.
          * strict_gap_mode = False then a gap has an increased tolerance. Some new system with different clock need this option
            otherwise, too many gaps are detected

    Notes
    -----
    * This IO supports NCS, NEV, NSE and NTT file formats (not NVT or NRD yet)

    * These variations of header format and possible differences between the stated sampling frequency
      and actual sampling frequency can create apparent time discrepancies in .Ncs files. Additionally,
      the Neuralynx recording software can start and stop recording while continuing to write samples
      to a single .Ncs file, which creates larger gaps in the time sequences of the samples.

    * This RawIO attempts to correct for these deviations where possible and present a single section of
      contiguous samples with one sampling frequency, t_start, and length for each .Ncs file. These
      sections are determined by the NcsSectionsFactory class. In the
      event the gaps are larger, this RawIO only provides the samples from the first section as belonging
      to one Segment.

    Examples
    --------
    >>> import neo.rawio
    >>> reader = neo.rawio.NeuralynxRawIO(dirname='Cheetah_v5.5.1/original_data')
    >>> reader.parse_header()

    Inspect all files in the directory.

    >>> print(reader)

     Display all information about signal channels, units, segment size....
    """

    extensions = ["nse", "ncs", "nev", "ntt", "nvt", "nrd"]  # nvt and nrd are not yet supported
    rawmode = "one-dir"

    _ncs_dtype = [
        ("timestamp", "uint64"),
        ("channel_id", "uint32"),
        ("sample_rate", "uint32"),
        ("nb_valid", "uint32"),
        ("samples", "int16", NcsSection._RECORD_SIZE),
    ]

    def __init__(
        self,
        dirname="",
        include_filenames=None,
        exclude_filenames=None,
        keep_original_times=False,
        strict_gap_mode=True,
        filename=None,
        exclude_filename=None,
        **kargs,
    ):

        if not dirname:
            raise ValueError("`dirname` cannot be empty.")

        if filename is not None:
            include_filenames = [filename]
            warnings.warn("`filename` is deprecated and will be removed. Please use `include_filenames` instead")

        if exclude_filename is not None:
            if isinstance(exclude_filename, str):
                exclude_filenames = [exclude_filename]
            else:
                exclude_filenames = exclude_filename
            warnings.warn(
                "`exclude_filename` is deprecated and will be removed. Please use `exclude_filenames` instead"
            )

        if include_filenames is None:
            include_filenames = []
        elif isinstance(include_filenames, str):
            include_filenames = [include_filenames]

        if exclude_filenames is None:
            exclude_filenames = []
        elif isinstance(exclude_filenames, str):
            exclude_filenames = [exclude_filenames]

        if include_filenames:
            self.rawmode = "multiple-files"
        else:
            self.rawmode = "one-dir"

        self.dirname = dirname
        self.include_filenames = include_filenames
        self.exclude_filenames = exclude_filenames
        self.keep_original_times = keep_original_times
        self.strict_gap_mode = strict_gap_mode
        BaseRawIO.__init__(self, **kargs)

    def _source_name(self):
        return self.dirname

    def _parse_header(self):

        stream_channels = []
        signal_channels = []
        spike_channels = []
        event_channels = []

        self.ncs_filenames = OrderedDict()  # (chan_name, chan_id): filename
        self.nse_ntt_filenames = OrderedDict()  # (chan_name, chan_id): filename
        self.nev_filenames = OrderedDict()  # chan_id: filename

        self.file_headers = OrderedDict()  # filename: file header dict

        self._nev_memmap = {}
        self._spike_memmap = {}
        self.internal_unit_ids = []  # channel_index > ((channel_name, channel_id), unit_id)
        self.internal_event_ids = []
        self._empty_ncs = []  # this list contains filenames of empty files
        self._empty_nev = []
        self._empty_nse_ntt = []

        # Explore the directory looking for ncs, nev, nse and ntt
        # and construct channels headers.
        signal_annotations = []
        unit_annotations = []
        event_annotations = []

        if self.rawmode == "one-dir":
            filenames = sorted(os.listdir(self.dirname))
        else:
            filenames = self.include_filenames

        filenames = [f for f in filenames if f not in self.exclude_filenames]
        full_filenames = [os.path.join(self.dirname, f) for f in filenames]

        for filename in full_filenames:
            if not os.path.isfile(filename):
                raise ValueError(
                    f"Provided Filename is not a file: "
                    f"{filename}. If you want to provide a "
                    f"directory use the `dirname` keyword"
                )

        stream_props = {}  # {(sampling_rate, n_samples, t_start): {stream_id: [filenames]}

        for filename in full_filenames:
            _, ext = os.path.splitext(filename)
            ext = ext[1:]  # remove dot
            ext = ext.lower()  # make lower case for comparisons
            if ext not in self.extensions:
                continue

            # Skip Ncs files with only header. Other empty file types
            # will have an empty dataset constructed later.
            if (os.path.getsize(filename) <= NlxHeader.HEADER_SIZE) and ext in ["ncs"]:
                self._empty_ncs.append(filename)
                continue

            # All file have more or less the same header structure
            info = NlxHeader(filename)
            self.file_headers[filename] = info
            chan_names = info["channel_names"]
            chan_ids = info["channel_ids"]

            for idx, chan_id in enumerate(chan_ids):
                chan_name = chan_names[idx]

                chan_uid = (chan_name, str(chan_id))
                if ext == "ncs":
                    file_mmap = self._get_file_map(filename)
                    n_packets = copy.copy(file_mmap.shape[0])
                    if n_packets:
                        t_start = copy.copy(file_mmap[0][0])
                    else:  # empty file
                        t_start = 0
                    stream_prop = (info["sampling_rate"], n_packets, t_start)
                    if stream_prop not in stream_props:
                        stream_props[stream_prop] = {"stream_id": len(stream_props), "filenames": [filename]}
                    else:
                        stream_props[stream_prop]["filenames"].append(filename)
                    stream_id = stream_props[stream_prop]["stream_id"]

                    # a sampled signal channel
                    units = "uV"
                    gain = info["bit_to_microVolt"][idx]
                    if info.get("input_inverted", False):
                        gain *= -1
                    offset = 0.0
                    signal_channels.append(
                        (chan_name, str(chan_id), info["sampling_rate"], "int16", units, gain, offset, stream_id)
                    )
                    self.ncs_filenames[chan_uid] = filename
                    keys = [
                        "DspFilterDelay_µs",
                        "recording_opened",
                        "FileType",
                        "DspDelayCompensation",
                        "recording_closed",
                        "DspLowCutFilterType",
                        "HardwareSubSystemName",
                        "DspLowCutNumTaps",
                        "DSPLowCutFilterEnabled",
                        "HardwareSubSystemType",
                        "DspHighCutNumTaps",
                        "ADMaxValue",
                        "DspLowCutFrequency",
                        "DSPHighCutFilterEnabled",
                        "RecordSize",
                        "InputRange",
                        "DspHighCutFrequency",
                        "input_inverted",
                        "NumADChannels",
                        "DspHighCutFilterType",
                    ]
                    d = {k: info[k] for k in keys if k in info}
                    signal_annotations.append(d)

                elif ext in ("nse", "ntt"):
                    # nse and ntt are pretty similar except for the waveform shape.
                    # A file can contain several unit_id (so several unit channel).
                    assert chan_id not in self.nse_ntt_filenames, "Several nse or ntt files have the same unit_id!!!"
                    self.nse_ntt_filenames[chan_uid] = filename

                    data = self._get_file_map(filename)
                    self._spike_memmap[chan_uid] = data

                    unit_ids = np.unique(data["unit_id"])
                    for unit_id in unit_ids:
                        # a spike channel for each (chan_id, unit_id)
                        self.internal_unit_ids.append((chan_uid, unit_id))

                        unit_name = "ch{}#{}#{}".format(chan_name, chan_id, unit_id)
                        unit_id = "{}".format(unit_id)
                        wf_units = "uV"
                        wf_gain = info["bit_to_microVolt"][idx]
                        if info.get("input_inverted", False):
                            wf_gain *= -1
                        wf_offset = 0.0
                        wf_left_sweep = -1  # NOT KNOWN
                        wf_sampling_rate = info["sampling_rate"]
                        spike_channels.append(
                            (
                                unit_name,
                                "{}".format(unit_id),
                                wf_units,
                                wf_gain,
                                wf_offset,
                                wf_left_sweep,
                                wf_sampling_rate,
                            )
                        )
                        unit_annotations.append(dict(file_origin=filename))

                elif ext == "nev":
                    # an event channel
                    # each ('event_id',  'ttl_input') give a new event channel
                    self.nev_filenames[chan_id] = filename

                    if os.path.getsize(filename) <= NlxHeader.HEADER_SIZE:
                        self._empty_nev.append(filename)
                        data = np.zeros((0,), dtype=nev_dtype)
                        internal_ids = []
                    else:
                        data = self._get_file_map(filename)
                        if data.shape[0] == 0:  # empty file
                            self._empty_nse_ntt.append(filename)
                        internal_ids = np.unique(data[["event_id", "ttl_input"]]).tolist()
                    for internal_event_id in internal_ids:
                        if internal_event_id not in self.internal_event_ids:
                            event_id, ttl_input = internal_event_id
                            name = "{} event_id={} ttl={}".format(chan_name, event_id, ttl_input)
                            event_channels.append((name, chan_id, "event"))
                            self.internal_event_ids.append(internal_event_id)

                    self._nev_memmap[chan_id] = data

        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        if signal_channels.size > 0:
            # ordering streams according from high to low sampling rates
            stream_props = {k: stream_props[k] for k in sorted(stream_props, reverse=True)}
            names = [f"Stream (rate,#packet,t0): {sp}" for sp in stream_props]
            ids = [stream_prop["stream_id"] for stream_prop in stream_props.values()]
            signal_streams = list(zip(names, ids))
        else:
            signal_streams = []
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        # set 2 attributes needed later for header in case there are no ncs files in dataset,
        #   e.g. Pegasus
        self._timestamp_limits = None
        self._nb_segment = 1

        stream_infos = {}

        # Read ncs files of each stream for gap detection and nb_segment computation.
        for stream_id in np.unique(signal_channels["stream_id"]):
            stream_channels = signal_channels[signal_channels["stream_id"] == stream_id]
            stream_chan_uids = zip(stream_channels["name"], stream_channels["id"])
            stream_filenames = [self.ncs_filenames[chuid] for chuid in stream_chan_uids]
            _sigs_memmaps, ncsSegTimestampLimits, section_structure = self.scan_stream_ncs_files(stream_filenames)

            stream_infos[stream_id] = {
                "segment_sig_memmaps": _sigs_memmaps,
                "ncs_segment_infos": ncsSegTimestampLimits,
                "section_structure": section_structure,
            }

        # check if section structure across streams is compatible and merge infos
        ref_sec_structure = None
        for stream_id, stream_info in stream_infos.items():
            ref_stream_id = list(stream_infos.keys())[0]
            ref_sec_structure = stream_infos[ref_stream_id]["section_structure"]

            sec_structure = stream_info["section_structure"]

            # check if section structure of streams are compatible
            # using tolerance of one data packet (512 samples)
            tolerance = 512 / min(ref_sec_structure.sampFreqUsed, sec_structure.sampFreqUsed) * 1e6
            if not ref_sec_structure.is_equivalent(sec_structure, abs_tol=tolerance):
                ref_chan_ids = signal_channels[signal_channels["stream_id"] == ref_stream_id]["name"]
                chan_ids = signal_channels[signal_channels["stream_id"] == stream_id]["name"]

                raise ValueError(
                    "Incompatible section structures across streams: "
                    f"Stream id {ref_stream_id}:{ref_chan_ids} and "
                    f"{stream_id}:{chan_ids}."
                )

        if ref_sec_structure is not None:
            self._nb_segment = len(ref_sec_structure.sects)
        else:
            # Use only a single segment if no ncs data is present
            self._nb_segment = 1

        def min_max_tuple(tuple1, tuple2):
            """Merge tuple by selecting min for first and max for 2nd entry"""
            mins, maxs = zip(tuple1, tuple2)
            result = (min(m for m in mins if m is not None), max(m for m in maxs if m is not None))
            return result

        # merge stream mmemmaps since streams are compatible
        self._sigs_memmaps = [{} for seg_idx in range(self._nb_segment)]
        # time limits of integer timestamps in ncs files
        self._timestamp_limits = [(None, None) for seg_idx in range(self._nb_segment)]
        # time limits physical times in ncs files
        self._signal_limits = [(None, None) for seg_idx in range(self._nb_segment)]
        for stream_id, stream_info in stream_infos.items():
            stream_mmaps = stream_info["segment_sig_memmaps"]
            for seg_idx, signal_dict in enumerate(stream_mmaps):
                self._sigs_memmaps[seg_idx].update(signal_dict)

            ncs_segment_info = stream_info["ncs_segment_infos"]
            for seg_idx, (t_start, t_stop) in enumerate(ncs_segment_info.timestamp_limits):
                self._timestamp_limits[seg_idx] = min_max_tuple(self._timestamp_limits[seg_idx], (t_start, t_stop))

            for seg_idx in range(ncs_segment_info.nb_segment):
                t_start = ncs_segment_info.t_start[seg_idx]
                t_stop = ncs_segment_info.t_stop[seg_idx]
                self._signal_limits[seg_idx] = min_max_tuple(self._signal_limits[seg_idx], (t_start, t_stop))

        # precompute signal lengths within segments
        self._sigs_length = []
        if self._sigs_memmaps:
            for seg_idx, sig_container in enumerate(self._sigs_memmaps):
                self._sigs_length.append({})
                for chan_uid, sig_infos in sig_container.items():
                    self._sigs_length[seg_idx][chan_uid] = int(sig_infos["nb_valid"].sum())

        # Determine timestamp limits in nev, nse, ntt files by scanning them.
        ts0, ts1 = None, None
        for _data_memmap in (self._spike_memmap, self._nev_memmap):
            for _, data in _data_memmap.items():
                ts = data["timestamp"]
                if ts.size == 0:
                    continue
                if ts0 is None:
                    ts0 = ts[0]
                    ts1 = ts[-1]
                ts0 = min(ts0, ts[0])
                ts1 = max(ts1, ts[-1])

        # rescaling for comparison with signal times
        if ts0 is not None:
            timestamps_start, timestamps_stop = ts0 / 1e6, ts1 / 1e6

        # decide on segment and global start and stop times based on files available
        if self._timestamp_limits is None:
            # case  NO ncs but HAVE nev or nse -> single segment covering all spikes & events
            self._timestamp_limits = [(ts0, ts1)]
            self._seg_t_starts = [timestamps_start]
            self._seg_t_stops = [timestamps_stop]
            self.global_t_start = timestamps_start
            self.global_t_stop = timestamps_stop
        elif ts0 is not None:
            # case  HAVE ncs AND HAVE nev or nse -> multi segments based on ncs segmentation
            # ignoring nev/nse/ntt time limits, loading only data within ncs segments
            global_events_limits = (timestamps_start, timestamps_stop)
            global_signal_limits = (self._signal_limits[0][0], self._signal_limits[-1][-1])
            self.global_t_start, self.global_t_stop = min_max_tuple(global_events_limits, global_signal_limits)
            self._seg_t_starts = [limits[0] for limits in self._signal_limits]
            self._seg_t_stops = [limits[1] for limits in self._signal_limits]
            self._seg_t_starts[0] = self.global_t_start
            self._seg_t_stops[-1] = self.global_t_stop

        else:
            # case HAVE ncs but  NO nev or nse ->
            self._seg_t_starts = [limits[0] for limits in self._signal_limits]
            self._seg_t_stops = [limits[1] for limits in self._signal_limits]
            self.global_t_start = self._signal_limits[0][0]
            self.global_t_stop = self._signal_limits[-1][-1]

        if self.keep_original_times:
            self.global_t_stop = self.global_t_stop - self.global_t_start
            self.global_t_start = 0

        # fill header dictionary
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [self._nb_segment]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # Annotations
        self._generate_minimal_annotations()
        bl_annotations = self.raw_annotations["blocks"][0]

        for seg_index in range(self._nb_segment):
            seg_annotations = bl_annotations["segments"][seg_index]

            for stream_id in range(signal_streams.size):
                # one or no signal stream
                stream_ann = seg_annotations["signals"][stream_id]
                # handle array annotations
                for key in signal_annotations[0].keys():
                    values = []
                    # only collect values from channels belonging to current stream
                    for d in np.where(signal_channels["stream_id"] == f"{stream_id}")[0]:
                        value = signal_annotations[d][key]
                        values.append(value)
                    values = np.array(values)
                    if values.ndim == 1:
                        # 'InputRange': is 2D and make bugs
                        stream_ann["__array_annotations__"][key] = values

            for c in range(spike_channels.size):
                unit_ann = seg_annotations["spikes"][c]
                unit_ann.update(unit_annotations[c])

            for c in range(event_channels.size):
                # annotations for channel events
                event_id, ttl_input = self.internal_event_ids[c]
                chan_id = event_channels[c]["id"]

                ev_ann = seg_annotations["events"][c]
                ev_ann["file_origin"] = self.nev_filenames[chan_id]

                # ~ ev_ann['marker_id'] =
                # ~ ev_ann['nttl'] =
                # ~ ev_ann['digital_marker'] =
                # ~ ev_ann['analog_marker'] =

    @staticmethod
    def _get_file_map(filename):
        """
        Create memory maps when needed
        see also https://github.com/numpy/numpy/issues/19340
        """
        filename = pathlib.Path(filename)
        suffix = filename.suffix.lower()[1:]

        if suffix == "ncs":
            return np.memmap(filename, dtype=NeuralynxRawIO._ncs_dtype, mode="r", offset=NlxHeader.HEADER_SIZE)

        elif suffix in ["nse", "ntt"]:
            info = NlxHeader(filename)
            dtype = get_nse_or_ntt_dtype(info, suffix)

            # return empty map if file does not contain data
            if os.path.getsize(filename) <= NlxHeader.HEADER_SIZE:
                return np.zeros((0,), dtype=dtype)

            return np.memmap(filename, dtype=dtype, mode="r", offset=NlxHeader.HEADER_SIZE)

        elif suffix == "nev":
            return np.memmap(filename, dtype=nev_dtype, mode="r", offset=NlxHeader.HEADER_SIZE)

        else:
            raise ValueError(f"Unknown file suffix {suffix}")

    # Accessors for segment times which are offset by appropriate global start time
    def _segment_t_start(self, block_index, seg_index):
        return self._seg_t_starts[seg_index] - self.global_t_start

    def _segment_t_stop(self, block_index, seg_index):
        return self._seg_t_stops[seg_index] - self.global_t_start

    def _get_signal_size(self, block_index, seg_index, stream_index):
        stream_id = self.header["signal_streams"][stream_index]["id"]
        stream_mask = self.header["signal_channels"]["stream_id"] == stream_id
        signals = self.header["signal_channels"][stream_mask]

        if len(signals):
            sig = signals[0]
            return self._sigs_length[seg_index][(sig["name"], sig["id"])]
        else:
            raise ValueError(
                f"No signals present for block {block_index}, segment {seg_index}," f" stream {stream_index}"
            )

    def _get_signal_t_start(self, block_index, seg_index, stream_index):

        stream_id = self.header["signal_streams"][stream_index]["id"]
        stream_mask = self.header["signal_channels"]["stream_id"] == stream_id

        # use first channel of stream as all channels in stream have a common t_start
        channel = self.header["signal_channels"][stream_mask][0]

        data = self._sigs_memmaps[seg_index][(channel["name"], channel["id"])]
        absolute_t_start = data["timestamp"][0]

        return absolute_t_start / 1e6 - self.global_t_start

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        """
        Retrieve chunk of analog signal, a chunk being a set of contiguous samples.

        Parameters
        ----------
        block_index:
            index of block in dataset, ignored as only 1 block in this implementation
        seg_index:
            index of segment to use
        i_start:
            sample index of first sample within segment to retrieve
        i_stop:
            sample index of last sample within segment to retrieve
        channel_indexes:
            list of channel indices to return data for

        Returns
        -------
            array of samples, with each requested channel in a column
        """
        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self.get_signal_size(block_index=block_index, seg_index=seg_index, stream_index=stream_index)

        block_start = i_start // NcsSection._RECORD_SIZE
        block_stop = i_stop // NcsSection._RECORD_SIZE + 1
        sl0 = i_start % 512
        sl1 = sl0 + (i_stop - i_start)

        if channel_indexes is None:
            channel_indexes = slice(None)

        stream_id = self.header["signal_streams"][stream_index]["id"]
        stream_mask = self.header["signal_channels"]["stream_id"] == stream_id

        channel_ids = self.header["signal_channels"][stream_mask][channel_indexes]["id"]
        channel_names = self.header["signal_channels"][stream_mask][channel_indexes]["name"]

        # create buffer for samples
        sigs_chunk = np.zeros((i_stop - i_start, len(channel_ids)), dtype="int16")

        for i, chan_uid in enumerate(zip(channel_names, channel_ids)):
            data = self._sigs_memmaps[seg_index][chan_uid]
            sub = data[block_start:block_stop]
            sigs_chunk[:, i] = sub["samples"].flatten()[sl0:sl1]

        return sigs_chunk

    def _spike_count(self, block_index, seg_index, unit_index):
        chan_uid, unit_id = self.internal_unit_ids[unit_index]
        data = self._spike_memmap[chan_uid]
        ts = data["timestamp"]

        ts0 = self.segment_t_start(block_index, seg_index)
        ts1 = self.segment_t_stop(block_index, seg_index)

        # rescale to integer sampling of time
        ts0 = int((ts0 + self.global_t_start) * 1e6)
        ts1 = int((ts1 + self.global_t_start) * 1e6)

        # only count spikes inside the timestamp limits, inclusive, and for the specified unit
        keep = (ts >= ts0) & (ts <= ts1) & (unit_id == data["unit_id"])
        nb_spike = int(data[keep].size)
        return nb_spike

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):
        """
        Extract timestamps within a Segment defined by ncs timestamps
        """
        chan_uid, unit_id = self.internal_unit_ids[unit_index]
        data = self._spike_memmap[chan_uid]
        ts = data["timestamp"]

        ts0, ts1 = t_start, t_stop
        if ts0 is None:
            ts0 = self.segment_t_start(block_index, seg_index)
        if ts1 is None:
            ts1 = self.segment_t_stop(block_index, seg_index)

        # rescale to integer sampling of time
        ts0 = int((ts0 + self.global_t_start) * 1e6)
        ts1 = int((ts1 + self.global_t_start) * 1e6)

        keep = (ts >= ts0) & (ts <= ts1) & (unit_id == data["unit_id"])
        timestamps = ts[keep]
        return timestamps

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        spike_times = spike_timestamps.astype(dtype)
        spike_times /= 1e6
        spike_times -= self.global_t_start
        return spike_times

    def _get_spike_raw_waveforms(self, block_index, seg_index, unit_index, t_start, t_stop):
        chan_uid, unit_id = self.internal_unit_ids[unit_index]
        data = self._spike_memmap[chan_uid]
        ts = data["timestamp"]

        ts0, ts1 = t_start, t_stop
        if ts0 is None:
            ts0 = self.segment_t_start(block_index, seg_index)
        if ts1 is None:
            ts1 = self.segment_t_stop(block_index, seg_index)

        # rescale to integer sampling of time
        ts0 = int((ts0 + self.global_t_start) * 1e6)
        ts1 = int((ts1 + self.global_t_start) * 1e6)

        keep = (ts >= ts0) & (ts <= ts1) & (unit_id == data["unit_id"])

        wfs = data[keep]["samples"]
        if wfs.ndim == 2:
            # case for nse
            waveforms = wfs[:, None, :]
        else:
            # case for ntt change (n, 32, 4) to (n, 4, 32)
            waveforms = wfs.swapaxes(1, 2)

        return waveforms

    def _event_count(self, block_index, seg_index, event_channel_index):
        event_id, ttl_input = self.internal_event_ids[event_channel_index]
        chan_id = self.header["event_channels"][event_channel_index]["id"]
        data = self._nev_memmap[chan_id]

        ts0 = self.segment_t_start(block_index, seg_index)
        ts1 = self.segment_t_stop(block_index, seg_index)

        # rescale to integer sampling of time
        ts0 = int((ts0 + self.global_t_start) * 1e6)
        ts1 = int((ts1 + self.global_t_start) * 1e6)

        ts = data["timestamp"]
        keep = (ts >= ts0) & (ts <= ts1) & (data["event_id"] == event_id) & (data["ttl_input"] == ttl_input)
        nb_event = int(data[keep].size)
        return nb_event

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        event_id, ttl_input = self.internal_event_ids[event_channel_index]
        chan_id = self.header["event_channels"][event_channel_index]["id"]
        data = self._nev_memmap[chan_id]

        ts0, ts1 = t_start, t_stop
        if ts0 is None:
            ts0 = self.segment_t_start(block_index, seg_index)
        if ts1 is None:
            ts1 = self.segment_t_stop(block_index, seg_index)

        # rescale to integer sampling of time
        ts0 = int((ts0 + self.global_t_start) * 1e6)
        ts1 = int((ts1 + self.global_t_start) * 1e6)

        ts = data["timestamp"]
        keep = (ts >= ts0) & (ts <= ts1) & (data["event_id"] == event_id) & (data["ttl_input"] == ttl_input)

        subdata = data[keep]
        timestamps = subdata["timestamp"]
        labels = subdata["event_string"].astype("U")
        durations = None
        return timestamps, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        event_times = event_timestamps.astype(dtype)
        event_times /= 1e6
        event_times -= self.global_t_start
        return event_times

    def scan_stream_ncs_files(self, ncs_filenames):
        """
        Given a list of ncs files, read their basic structure.
        Ncs files have to have common sampling_rate, number of packets and t_start
        (be part of a single stream)

        Parameters
        ----------
        ncs_filenames: list
            List of ncs filenames to scan.

        Returns
        -------
        memmaps
            [ {} for seg_index in range(self._nb_segment) ][chan_uid]
        seg_time_limits
            SegmentTimeLimits for sections in scanned Ncs files
        section_structure
            Section structure common to the ncs files

        Files will be scanned to determine the sections of records. If file is a single
        section of records, this scan is brief, otherwise it will check each record which may
        take some time.
        """

        if len(ncs_filenames) == 0:
            return None, None, None

        # Build dictionary of chan_uid to associated NcsSections, memmap and NlxHeaders. Only
        # construct new NcsSections when it is different from that for the preceding file.
        chanSectMap = dict()
        sig_length = []
        for ncs_filename in ncs_filenames:

            data = self._get_file_map(ncs_filename)
            nlxHeader = NlxHeader(ncs_filename)

            verify_sec_struct = NcsSectionsFactory._verifySectionsStructure
            if not chanSectMap or (not verify_sec_struct(data, chan_ncs_sections)):
                chan_ncs_sections = NcsSectionsFactory.build_for_ncs_file(
                    data, nlxHeader, strict_gap_mode=self.strict_gap_mode
                )

            # register file section structure for all contained channels
            for chan_uid in zip(nlxHeader["channel_names"], np.asarray(nlxHeader["channel_ids"], dtype=str)):
                chanSectMap[chan_uid] = [chan_ncs_sections, nlxHeader, ncs_filename]

            del data

        # Construct an inverse dictionary from NcsSections to list of associated chan_uids
        revSectMap = dict()
        for k, v in chanSectMap.items():
            revSectMap.setdefault(v[0], []).append(k)

        # If there is only one NcsSections structure in the set of ncs files, there should only
        # be one entry. Otherwise this is presently unsupported.
        if len(revSectMap) > 1:
            raise IOError(
                f"ncs files have {len(revSectMap)} different sections "
                f"structures. Unsupported configuration to be handled with in a single "
                f"stream."
            )

        seg_time_limits = SegmentTimeLimits(
            nb_segment=len(chan_ncs_sections.sects), t_start=[], t_stop=[], length=[], timestamp_limits=[]
        )
        memmaps = [{} for seg_index in range(seg_time_limits.nb_segment)]

        # create segment with subdata block/t_start/t_stop/length for each channel
        for i, chan_uid in enumerate(chanSectMap.keys()):
            data = self._get_file_map(chanSectMap[chan_uid][2])

            # create a memmap for each record section of the current file
            curSects = chanSectMap[chan_uid][0]
            for seg_index in range(len(curSects.sects)):

                curSect = curSects.sects[seg_index]
                subdata = data[curSect.startRec : (curSect.endRec + 1)]
                memmaps[seg_index][chan_uid] = subdata

                # create segment timestamp limits based on only NcsSections structure in use
                if i == 0:
                    numSampsLastSect = subdata[-1]["nb_valid"]
                    ts0 = subdata[0]["timestamp"]
                    ts1 = NcsSectionsFactory.calc_sample_time(
                        curSects.sampFreqUsed, subdata[-1]["timestamp"], numSampsLastSect
                    )
                    seg_time_limits.timestamp_limits.append((ts0, ts1))
                    t_start = ts0 / 1e6
                    seg_time_limits.t_start.append(t_start)
                    t_stop = ts1 / 1e6
                    seg_time_limits.t_stop.append(t_stop)
                    # :NOTE: This should really be the total of nb_valid in records, but this
                    #  allows the last record of a section to be shorter, the most common case.
                    #  Have never seen a section of records with not full records before the last.
                    length = (subdata.size - 1) * NcsSection._RECORD_SIZE + numSampsLastSect
                    seg_time_limits.length.append(length)

        stream_section_structure = list(revSectMap.keys())[0]

        return memmaps, seg_time_limits, stream_section_structure


# time limits for set of segments
SegmentTimeLimits = namedtuple("SegmentTimeLimits", ["nb_segment", "t_start", "t_stop", "length", "timestamp_limits"])

nev_dtype = [
    ("reserved", "<i2"),
    ("system_id", "<i2"),
    ("data_size", "<i2"),
    ("timestamp", "<u8"),
    ("event_id", "<i2"),
    ("ttl_input", "<i2"),
    ("crc_check", "<i2"),
    ("dummy1", "<i2"),
    ("dummy2", "<i2"),
    ("extra", "<i4", (8,)),
    ("event_string", "S128"),
]


def get_nse_or_ntt_dtype(info, ext):
    """
    For NSE and NTT the dtype depend on the header.

    """
    dtype = [("timestamp", "uint64"), ("channel_id", "uint32"), ("unit_id", "uint32")]

    # for purpose of dtypes, features in the file are always fixed 8 presently,
    # whether mentioned in the header or not. Features may not be listed in the header
    # if no feature names are assigned in Neuralynx software.
    nb_feature = 8
    dtype += [("features", "int32", (nb_feature,))]

    # Number of samples are fixed in the file at 32 for .nse 32 * 4 for .ntt.
    # WaveformLength may or may not be listed in the file depending on settings
    # in the Neuralynx software, so don't try retrieving it.
    if ext == "nse":
        nb_sample = 32
        dtype += [("samples", "int16", (nb_sample,))]
    elif ext == "ntt":
        nb_sample = 32
        nb_chan = 4
        dtype += [("samples", "int16", (nb_sample, nb_chan))]

    return dtype



"""
Class for reading data from Neuralynx files.
This IO supports NCS, NEV and NSE file formats.

Depends on: numpy

Supported: Read

Author: Julia Sprenger, Carlos Canova
"""

import warnings
from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.neuralynxrawio.neuralynxrawio import NeuralynxRawIO


class NeuralynxIO(NeuralynxRawIO, BaseFromRaw):
    """
    Class for reading data from Neuralynx files.
    This IO supports NCS, NEV, NSE and NTT file formats.

    NCS contains signals for one channel
    NEV contains events
    NSE contains spikes and waveforms for mono electrodes
    NTT contains spikes and waveforms for tetrodes
    """

    _prefered_signal_group_mode = "group-by-same-units"
    mode = "dir"

    def __init__(
        self,
        dirname,
        use_cache=False,
        cache_path="same_as_resource",
        include_filenames=None,
        exclude_filenames=None,
        keep_original_times=False,
        filename=None,
        exclude_filename=None,
    ):
        """
        Initialise IO instance

        Parameters
        ----------
        dirname : str
            Directory containing data files
        filename : str
            Deprecated and will be removed. Please use `include_filenames` instead
            Name of a single ncs, nse, nev, or ntt file to include in dataset. Will be ignored,
            if dirname is provided.
        use_cache : bool, optional
            Cache results of initial file scans for faster loading in subsequent runs.
            Default: False
        cache_path : str, optional
            Folder path to use for cache files.
            Default: 'same_as_resource'
        exclude_filename: None,
            Deprecated and will be removed. Please use `exclude_filenames` instead
        include_filenames: str or list
            Filename or list of filenames to be included. This can be absolute path or path relative to dirname.
        exclude_filenames: str or list
            Filename or list of filenames to be excluded. Expects base filenames without
            directory path.
        keep_original_times : bool
            Preserve original time stamps as in data files. By default datasets are
            shifted to begin at t_start = 0*pq.second.
            Default: False
        """

        if filename is not None:
            warnings.warn("Deprecated and will be removed. Please use `include_filenames` instead")
            include_filenames = [filename]

        if exclude_filename is not None:
            warnings.warn("Deprecated and will be removed. Please use `exclude_filenames` instead")
            exclude_filenames = exclude_filename

        NeuralynxRawIO.__init__(
            self,
            dirname=dirname,
            include_filenames=include_filenames,
            exclude_filenames=exclude_filenames,
            keep_original_times=keep_original_times,
            use_cache=use_cache,
            cache_path=cache_path,
        )

        if self.rawmode == "one-dir":
            BaseFromRaw.__init__(self, dirname)
        elif self.rawmode == "multiple-files":
            BaseFromRaw.__init__(self, include_filenames=include_filenames)



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.neuroscoperawio import NeuroScopeRawIO


class NeuroScopeIO(NeuroScopeRawIO, BaseFromRaw):
    """
    Reading from Neuroscope format files.

    Ref: http://neuroscope.sourceforge.net/
    """

    _prefered_signal_group_mode = "group-by-same-units"

    def __init__(self, filename):
        NeuroScopeRawIO.__init__(self, filename=filename)
        BaseFromRaw.__init__(self, filename)



"""
Tests of neo.io.blackrockio
"""

import unittest
import warnings

from numpy.testing import assert_equal

import numpy as np
import quantities as pq

from neo.io.blackrockio import BlackrockIO
from neo.core import NeoReadWriteError

from neo.test.iotest.common_io_test import BaseTestIO
from neo.test.tools import assert_neo_object_is_compliant
from neo.test.rawiotest.test_blackrockrawio import TestBlackrockRawIO

# check scipy
try:
    from packaging.version import Version
    import scipy.io
    import scipy.version
except ImportError as err:
    HAVE_SCIPY = False
    SCIPY_ERR = err
else:
    if Version(scipy.version.version) < Version("0.8"):
        HAVE_SCIPY = False
        SCIPY_ERR = ImportError(
            "your scipy version is too old to support "
            + "MatlabIO, you need at least 0.8. "
            + "You have %s" % scipy.version.version
        )
    else:
        HAVE_SCIPY = True
        SCIPY_ERR = None


class CommonTests(BaseTestIO, unittest.TestCase):
    ioclass = BlackrockIO
    entities_to_download = ["blackrock"]
    entities_to_test = [
        "blackrock/FileSpec2.3001",
        "blackrock/blackrock_2_1/l101210-001",
        "blackrock/blackrock_3_0/file_spec_3_0",
    ]

    def test_load_waveforms(self):
        filename = self.get_local_path("blackrock/FileSpec2.3001")
        reader = BlackrockIO(filename=filename, verbose=False)

        bl = reader.read_block(load_waveforms=True)
        assert_neo_object_is_compliant(bl)

    def test_inputs_V23(self):
        """
        Test various inputs to BlackrockIO.read_block with version 2.3 file
        to check for parsing errors.
        """
        filename = self.get_local_path("blackrock/FileSpec2.3001")
        reader = BlackrockIO(filename=filename, verbose=False, nsx_to_load=5)

        # Assert IOError is raised when no Blackrock files are available
        with self.assertRaises(IOError):
            reader2 = BlackrockIO(filename="nonexistent")

        # Load data to maximum extent, one None is not given as list
        block = reader.read_block(load_waveforms=False)
        lena = len(block.segments[0].analogsignals[0])
        numspa = len(block.segments[0].spiketrains[0])

        # Load data using a negative time and a time exceeding the end of the
        # recording
        too_large_tstop = block.segments[0].analogsignals[0].t_stop + 1 * pq.s
        buggy_slice = (-100 * pq.ms, too_large_tstop)

        # this is valid in read_segment
        seg = reader.read_segment(seg_index=0, time_slice=buggy_slice, strict_slicing=False)
        # this raise an error
        with self.assertRaises(ValueError):
            seg = reader.read_segment(seg_index=0, time_slice=buggy_slice, strict_slicing=True)

        lenb = len(seg.analogsignals[0])
        numspb = len(seg.spiketrains[0])

        # Same length of analog signal?
        # Both should have read the complete data set!
        self.assertEqual(lena, lenb)

        # Same length of spike train?
        # Both should have read the complete data set!
        self.assertEqual(numspa, numspb)

        # test 4 Units
        block = reader.read_block(load_waveforms=True, signal_group_mode="split-all")

        self.assertEqual(len(block.segments[0].analogsignals), 10)
        # self.assertEqual(len(block.channel_indexes[-1].units), 4)
        # self.assertEqual(len(block.channel_indexes[-1].units),
        #                 len(block.segments[0].spiketrains))

        anasig = block.segments[0].analogsignals[0]
        self.assertIsNotNone(anasig.file_origin)

    def test_inputs_V21(self):
        """
        Test various inputs to BlackrockIO.read_block with version 2.3 file
        to check for parsing errors.
        """
        filename = self.get_local_path("blackrock/blackrock_2_1/l101210-001")
        reader = BlackrockIO(filename=filename, verbose=False, nsx_to_load=5)

        # Assert IOError is raised when no Blackrock files are available
        with self.assertRaises(IOError):
            reader2 = BlackrockIO(filename="nonexistent")
        # with self.assertRaises(IOError):
        #     reader2 = BlackrockIO(filename=filename, nev_override='nonexistent')

        # Load data to maximum extent, one None is not given as list
        block = reader.read_block(load_waveforms=False, signal_group_mode="split-all")
        lena = len(block.segments[0].analogsignals[0])
        numspa = len(block.segments[0].spiketrains[0])

        # Load data using a negative time and a time exceeding the end of the
        # recording
        too_large_tstop = block.segments[0].analogsignals[0].t_stop + 1 * pq.s
        buggy_slice = (-100 * pq.ms, too_large_tstop)

        # This is valid in read_segment
        seg = reader.read_segment(seg_index=0, time_slice=buggy_slice, strict_slicing=False)
        # this raise error
        with self.assertRaises(ValueError):
            seg = reader.read_segment(seg_index=0, time_slice=buggy_slice, strict_slicing=True)

        lenb = len(seg.analogsignals[0])
        numspb = len(seg.spiketrains[0])

        # Same length of analog signal?
        # Both should have read the complete data set!
        self.assertEqual(lena, lenb)

        # Same length of spike train?
        # Both should have read the complete data set!
        self.assertEqual(numspa, numspb)

        # test 4 Units
        block = reader.read_block(
            load_waveforms=True,
            signal_group_mode="split-all",
        )  # units_group_mode='all-in-one')

        self.assertEqual(len(block.segments[0].analogsignals), 96)
        # self.assertEqual(len(block.channel_indexes[-1].units), 218)
        # self.assertEqual(len(block.channel_indexes[-1].units),
        #                 len(block.segments[0].spiketrains))
        anasig = block.segments[0].analogsignals[0]
        self.assertIsNotNone(anasig.file_origin)

    def test_load_muliple_nsx(self):
        """
        Test if multiple nsx signals can be loaded at the same time.
        """
        filename = self.get_local_path("blackrock/blackrock_2_1/l101210-001")
        reader = BlackrockIO(filename=filename, verbose=False, nsx_to_load="all")

        # number of different sampling rates corresponds to number of nsx signals, because
        # single nsx contains only signals of identical sampling rate
        block = reader.read_block(load_waveforms=False)
        sampling_rates = np.unique([a.sampling_rate.rescale("Hz") for a in block.filter(objects="AnalogSignal")])
        self.assertEqual(len(sampling_rates), 2)

        segment = reader.read_segment()
        sampling_rates = np.unique([a.sampling_rate.rescale("Hz") for a in segment.filter(objects="AnalogSignal")])
        self.assertEqual(len(sampling_rates), 2)

        # load only ns5
        reader = BlackrockIO(filename=filename, nsx_to_load=5)
        seg = reader.read_segment()
        self.assertEqual(len(seg.analogsignals), 1)
        self.assertEqual(seg.analogsignals[0].shape, (109224, 96))

        # load only ns2
        reader = BlackrockIO(filename=filename, nsx_to_load=2)
        seg = reader.read_segment()
        self.assertEqual(len(seg.analogsignals), 1)
        self.assertEqual(seg.analogsignals[0].shape, (3640, 6))

        # load only ns2
        reader = BlackrockIO(filename=filename, nsx_to_load=[2])
        seg = reader.read_segment()
        self.assertEqual(len(seg.analogsignals), 1)

        # load ns2 + ns5
        reader = BlackrockIO(filename=filename, nsx_to_load=[2, 5])
        seg = reader.read_segment()
        self.assertEqual(len(seg.analogsignals), 2)
        self.assertEqual(seg.analogsignals[0].shape, (3640, 6))
        self.assertEqual(seg.analogsignals[1].shape, (109224, 96))

        # load only ns5
        reader = BlackrockIO(filename=filename, nsx_to_load="max")
        seg = reader.read_segment()
        self.assertEqual(len(seg.analogsignals), 1)
        self.assertEqual(seg.analogsignals[0].shape, (109224, 96))

    @unittest.skipUnless(HAVE_SCIPY, "requires scipy")
    def test_compare_blackrockio_with_matlabloader_v21(self):
        """
        This test compares the output of BlackrockIO.read_block() with the
        output generated by a Matlab implementation of a Blackrock file reader
        provided by the company. The output for comparison is provided in a
        .mat file created by the script create_data_matlab_blackrock.m.
        The function tests LFPs, spike times, and digital events.
        """

        dirname = self.get_local_path("blackrock/blackrock_2_1/l101210-001")
        # First run with parameters for ns5, then run with correct parameters for ns2
        parameters = [
            (
                "blackrock/blackrock_2_1/l101210-001_nev-02_ns5.mat",
                {"nsx_to_load": 5, "nev_override": "-".join([dirname, "02"])},
            ),
            ("blackrock/blackrock_2_1/l101210-001.mat", {"nsx_to_load": 2}),
        ]
        for index, param in enumerate(parameters):
            # Load data from matlab generated files
            ml = scipy.io.loadmat(self.get_local_path(param[0]))
            lfp_ml = ml["lfp"]  # (channel x time) LFP matrix
            ts_ml = ml["ts"]  # spike time stamps
            elec_ml = ml["el"]  # spike electrodes
            unit_ml = ml["un"]  # spike unit IDs
            wf_ml = ml["wf"]  # waveforms
            mts_ml = ml["mts"]  # marker time stamps
            mid_ml = ml["mid"]  # marker IDs

            # Load data from original data files using the Neo BlackrockIO
            session = BlackrockIO(dirname, verbose=False, **param[1])
            block = session.read_block(load_waveforms=True, signal_group_mode="split-all")
            # Check if analog data are equal
            self.assertGreater(len(block.groups), 0)

            # Check if spikes are equal
            self.assertEqual(len(block.segments), 1)
            for st_i in block.segments[0].spiketrains:
                channelid = st_i.annotations["channel_id"]
                unitid = st_i.annotations["unit_id"]

                # Compare waveforms
                matlab_wf = wf_ml[np.nonzero(np.logical_and(elec_ml == channelid, unit_ml == unitid)), :][0]
                # Atleast_2d as correction for waveforms that are saved
                # in single dimension in SpikeTrain
                # because only one waveform is available
                assert_equal(np.atleast_2d(np.squeeze(st_i.waveforms).magnitude), matlab_wf)

                # Compare spike timestamps
                matlab_spikes = ts_ml[np.nonzero(np.logical_and(elec_ml == channelid, unit_ml == unitid))]
                # Going sure that unit is really seconds and not 1/30000 seconds
                if (not st_i.units == pq.CompoundUnit("1.0/{} * s".format(30000))) and st_i.units == pq.s:
                    st_i = np.round(st_i.base * 30000).astype(int)
                assert_equal(st_i, matlab_spikes)

            # Check if digital input port events are equal
            self.assertGreater(len(block.segments[0].events), 0)
            for ea_i in block.segments[0].events:
                if ea_i.name == "digital_input_port":
                    # Get all digital event IDs in this recording
                    marker_ids = set(ea_i.labels)
                    for marker_id in marker_ids:
                        python_digievents = np.round(ea_i.times.base[ea_i.labels == marker_id] * 30000).astype(int)
                        matlab_digievents = mts_ml[np.nonzero(mid_ml == int(marker_id))]
                        assert_equal(python_digievents, matlab_digievents)

                        # Note: analog input events are not yet supported

    def test_segment_detection_reset(self):
        """
        This test makes sure segments are detected correctly when reset was used during recording.
        """

        # Path to nev that will fail
        filename_nev_fail = self.get_local_path("blackrock/segment/ResetFail/reset_fail")
        # Path to nsX and nev that will NOT fail
        filename = self.get_local_path("blackrock/segment/ResetCorrect/reset")

        # Warning filter needs to be set to always before first occurrence of this warning
        warnings.simplefilter("always", UserWarning)

        # This fails, because in the nev there is no way to separate two segments
        with self.assertRaises(NeoReadWriteError):
            reader = BlackrockIO(filename=filename, nsx_to_load=2, nev_override=filename_nev_fail)

        # The correct file will issue a warning because a reset has occurred
        # and could be detected, but was not explicitly documented in the file
        with warnings.catch_warnings(record=True) as w:
            reader = BlackrockIO(filename=filename, nsx_to_load=2)
            self.assertGreaterEqual(len(w), 1)
            messages = [str(warning.message) for warning in w if warning.category == UserWarning]
            self.assertIn("Detected 1 undocumented segments within nev data after " "timestamps [5451].", messages)

        # Manually reset warning filter in order to not show too many warnings afterwards
        warnings.simplefilter("default")

        block = reader.read_block(load_waveforms=False, signal_group_mode="split-all")

        # 1 Segment at the beginning and 1 after reset
        self.assertEqual(len(block.segments), 2)
        # Checking all times are correct as read from file itself
        # (taking neo calculations into account)
        self.assertEqual(block.segments[0].t_start, 0.0)
        self.assertEqual(block.segments[0].t_stop, 4.02)
        # Clock is reset to 0
        self.assertEqual(block.segments[1].t_start, 0.0032)
        self.assertEqual(block.segments[1].t_stop, 3.9842)
        self.assertEqual(block.segments[0].analogsignals[0].t_start, 0.0)
        self.assertEqual(block.segments[0].analogsignals[0].t_stop, 4.02)
        self.assertEqual(block.segments[1].analogsignals[0].t_start, 0.0032)
        self.assertEqual(block.segments[1].analogsignals[0].t_stop, 3.9842)
        self.assertEqual(block.segments[0].spiketrains[0].t_start, 0.0)
        self.assertEqual(block.segments[0].spiketrains[0].t_stop, 4.02)
        self.assertEqual(block.segments[1].spiketrains[0].t_start, 0.0032)
        self.assertEqual(block.segments[1].spiketrains[0].t_stop, 3.9842)

        # Each segment must have the same number of analogsignals
        self.assertEqual(len(block.segments[0].analogsignals), len(block.segments[1].analogsignals))

        # Length of analogsignals as created
        self.assertEqual(len(block.segments[0].analogsignals[0][:]), 4020)
        self.assertEqual(len(block.segments[1].analogsignals[0][:]), 3981)

    def test_segment_detection_pause(self):
        """
        This test makes sure segments are detected correctly when pause was used during recording.
        """

        # Path to nev that has spikes that don't fit nsX segment
        filename_nev_outside_seg = self.get_local_path("blackrock/segment/PauseSpikesOutside/pause_spikes_outside_seg")
        # Path to nsX and nev that are correct
        filename = self.get_local_path("blackrock/segment/PauseCorrect/pause_correct")

        # This issues a warning, because there are spikes a long time after the last segment
        # And another one because there are spikes between segments
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            reader = BlackrockIO(filename=filename, nsx_to_load=2, nev_override=filename_nev_outside_seg)
            self.assertGreaterEqual(len(w), 2)

            # Check that warnings are correct
            messages = [str(warning.message) for warning in w if warning.category == UserWarning]
            self.assertIn("Spikes outside any segment. Detected on segment #1", messages)
            self.assertIn("Spikes 0.0776s after last segment.", messages)

        block = reader.read_block(load_waveforms=False, signal_group_mode="split-all")

        # 2 segments
        self.assertEqual(len(block.segments), 2)

        # Checking all times are correct as read from file itself
        # (taking neo calculations into account)
        self.assertEqual(block.segments[0].t_start, 0.0)
        # This value is so high, because a spike occurred right before the second segment
        # And thus is added to the first segment
        # This is not normal behavior and occurs because of the way the files were cut
        # into test files
        self.assertAlmostEqual(block.segments[0].t_stop.magnitude, 15.83916667)
        # Clock is not reset
        self.assertEqual(block.segments[1].t_start.magnitude, 31.0087)
        # Segment time is longer here as well because of spikes after second segment
        self.assertEqual(block.segments[1].t_stop.magnitude, 35.0863)
        self.assertEqual(block.segments[0].analogsignals[0].t_start, 0.0)
        # The AnalogSignal is only 4 seconds long, as opposed to the segment
        # whose length is caused by the additional spike
        self.assertEqual(block.segments[0].analogsignals[0].t_stop, 4.0)
        self.assertEqual(block.segments[1].analogsignals[0].t_start, 31.0087)
        self.assertAlmostEqual(block.segments[1].analogsignals[0].t_stop.magnitude, 35.0087, places=6)
        self.assertEqual(block.segments[0].spiketrains[0].t_start, 0.0)
        self.assertAlmostEqual(block.segments[0].spiketrains[0].t_stop.magnitude, 15.83916667, places=8)
        self.assertEqual(block.segments[1].spiketrains[0].t_start, 31.0087)
        self.assertEqual(block.segments[1].spiketrains[0].t_stop, 35.0863)

        # Each segment has same number of analogsignals
        self.assertEqual(len(block.segments[0].analogsignals), len(block.segments[1].analogsignals))

        # Analogsignals have exactly 4000 samples
        self.assertEqual(len(block.segments[0].analogsignals[0][:]), 4000)
        self.assertEqual(len(block.segments[1].analogsignals[0][:]), 4000)

        # This case is correct, no spikes outside segment or anything
        reader = BlackrockIO(filename=filename, nsx_to_load=2)
        block = reader.read_block(load_waveforms=False, signal_group_mode="split-all")

        # 2 segments
        self.assertEqual(len(block.segments), 2)

        # Checking all times are correct as read from file itself
        # (taking neo calculations into account)
        self.assertEqual(block.segments[0].t_start, 0.0)
        # Now segment time is only 4 seconds, because there were no additional spikes
        self.assertEqual(block.segments[0].t_stop, 4.0)
        self.assertEqual(block.segments[1].t_start, 31.0087)
        self.assertAlmostEqual(block.segments[1].t_stop.magnitude, 35.0087, places=6)
        self.assertEqual(block.segments[0].analogsignals[0].t_start, 0.0)
        self.assertEqual(block.segments[0].analogsignals[0].t_stop, 4.0)
        self.assertEqual(block.segments[1].analogsignals[0].t_start, 31.0087)
        self.assertAlmostEqual(block.segments[1].analogsignals[0].t_stop.magnitude, 35.0087, places=6)
        self.assertEqual(block.segments[0].spiketrains[0].t_start, 0.0)
        self.assertEqual(block.segments[0].spiketrains[0].t_stop, 4.0)
        self.assertEqual(block.segments[1].spiketrains[0].t_start, 31.0087)
        self.assertAlmostEqual(block.segments[1].spiketrains[0].t_stop.magnitude, 35.0087, places=6)

        # Each segment has same number of analogsignals
        self.assertEqual(len(block.segments[0].analogsignals), len(block.segments[1].analogsignals))

        # ns2 was created in such a way that all analogsignals have 4000 samples
        self.assertEqual(len(block.segments[0].analogsignals[0][:]), 4000)
        self.assertEqual(len(block.segments[1].analogsignals[0][:]), 4000)


if __name__ == "__main__":
    unittest.main()



"""
Reading files with neo.rawio
============================

compare with read_files_neo_io.py
"""

###########################################################
# First we import a RawIO from neo.rawio
# For this example we will use PlexonRawIO

import urllib
from neo.rawio import PlexonRawIO

##############################################################
# Get Plexon files

# We will be pulling these files down, but if you have a local file
# then all you need to do is specify the file location on your
# computer. NeuralEnsemble keeps a wide variety of freely accesible, small
# test files that can be used. So for this example we will take advantage of that
# fact.

url_repo = "https://web.gin.g-node.org/NeuralEnsemble/ephy_testing_data/raw/master/"
distantfile = url_repo + "plexon/File_plexon_3.plx"
localfile = "File_plexon_3.plx"
urllib.request.urlretrieve(distantfile, localfile)

###############################################################
# Create a reader

# All it takes to create a reader is giving the filename (or dirname)
# Then we need to do the slow step of parsing the header with the
# `parse_header` function. This collects metadata as well as
# make all future steps much faster for us

reader = PlexonRawIO(filename="File_plexon_3.plx")
reader.parse_header()
print(reader)
# we can view metadata in the header
print(reader.header)

###############################################################
# Read signal chunks
# This is how we read raw data. We choose indices that we want or
# we can use None to mean look at all channels. We also need to
# specify the block of data (block_index) as well as the segment
# (seg_index). Then we give the index start and stop. Since we
# often think in time: to go from time to index would just require
# the sample rate (so index = time / sampling_rate)

channel_indexes = None  # could be channel_indexes = [0]
raw_sigs = reader.get_analogsignal_chunk(
    block_index=0, seg_index=0, i_start=1024, i_stop=2048, channel_indexes=channel_indexes
)

# raw_sigs are not voltages so to convert to voltages we do the follwing
float_sigs = reader.rescale_signal_raw_to_float(raw_sigs, dtype="float64")

# We can see that the shapes are the same, but that the datatypes
# are different once we've rescaled our data
print("Raw Data: ", raw_sigs.shape, raw_sigs.dtype)
print("Scaled Data: ", float_sigs.shape, float_sigs.dtype)

###############################################################
# Each rawio gives you access to lots of information about your data
# some of this information comes from functions
# other information is stored as metadata in the reader.header

sampling_rate = reader.get_signal_sampling_rate()
# Like above we need to indicate the block and segment
t_start = reader.get_signal_t_start(block_index=0, seg_index=0)
units = reader.header["signal_channels"][0]["units"]

# and we can display all of this information
print(f"{sampling_rate=}, {t_start=}, {units=}")


####################################################################
# Some rawio's and file formats also provide information about spikes
# If a rawio can't read this data it will raise an error, but lucky
# for us PlexonRawIO does have spikes data!!

# Count units and spikes per unit
nb_unit = reader.spike_channels_count()
print(f"nb_unit: {nb_unit}\n")  # nb_unit stands for number of units
print("spike_channel_index     nb_spike")
for spike_channel_index in range(nb_unit):
    nb_spike = reader.spike_count(block_index=0, seg_index=0, spike_channel_index=spike_channel_index)
    print(f"{spike_channel_index}: {nb_spike}\n")

# Read spike times and rescale (just like analogsignal above)
spike_timestamps = reader.get_spike_timestamps(
    block_index=0, seg_index=0, spike_channel_index=0, t_start=0.0, t_stop=10.0
)

print(f"{spike_timestamps.shape=}\n{spike_timestamps.dtype=}\n{spike_timestamps[:5]=}\n")
spike_times = reader.rescale_spike_timestamp(spike_timestamps, dtype="float64")
print(f"{spike_times.shape=}\n{spike_times.dtype=}\n{spike_times[:5]}\n")

#######################################################################
# Some file formats can also give waveform information. We are lucky
# again our file has waveform data!! We forms are a 3d dataset of
# (nb_spike, nb_channel, nb_sample)

# Read spike waveforms
raw_waveforms = reader.get_spike_raw_waveforms(
    block_index=0, seg_index=0, spike_channel_index=0, t_start=0.0, t_stop=10.0
)
print(f"{raw_waveforms.shape=}\n{raw_waveforms.dtype=}\n{raw_waveforms[0, 0, :4]=}\n")
float_waveforms = reader.rescale_waveforms_to_float(raw_waveforms, dtype="float32", spike_channel_index=0)
print(f"{float_waveforms.shape=}\n{float_waveforms.dtype=}{float_waveforms[0, 0, :4]=}\n")

#########################################################################
# RawIOs can also read event timestamps. But looks like our luck ran out
# let's grab a new file to see this feature

# Read event timestamps and times (take another file)
distantfile = url_repo + "plexon/File_plexon_2.plx"
localfile = "File_plexon_2.plx"
urllib.request.urlretrieve(distantfile, localfile)

#########################################################################
# Since this is a new file we need to read initialize our reader as well
# as parse the header

reader = PlexonRawIO(filename="File_plexon_2.plx")
reader.parse_header()
# if we look at this header we see it is different than the header above
print(reader.header)

###########################################################################
# Now let's look at some event data. This could be things like stimuli applied
# during the course of an ephys recording

nb_event_channel = reader.event_channels_count()
print(f"nb_event_channel: {nb_event_channel}")
# now iterate through the channels
for chan_index in range(nb_event_channel):
    nb_event = reader.event_count(block_index=0, seg_index=0, event_channel_index=chan_index)
    print(f"chan_index: {chan_index} nb_event: {nb_event}\n")


###############################################################################
# Finally we can get our actual event timestamps. Some file formats provide the
# real timestamps (timestamps in s/ms) others have raw timestamps (in samples)
# so we can do the same style of functions as above. Get the raw timestamps
# and convert to real times with a rescale function.

ev_timestamps, ev_durations, ev_labels = reader.get_event_timestamps(
    block_index=0, seg_index=0, event_channel_index=0, t_start=None, t_stop=None
)
print(f"{ev_timestamps=}\n{ev_durations=}\n{ev_labels=}\n")
ev_times = reader.rescale_event_timestamp(ev_timestamps, dtype="float64")
print(f"{ev_times=}\n")



"""

Support for intan tech rhd and rhs files.

These 2 formats are more or less the same but:
  * some variance in headers.
  * rhs amplifier is more complex because the optional DC channel

RHS supported version 1.0
RHD supported version  1.0 1.1 1.2 1.3 2.0 3.0, 3.1
RHD headerless binary support 3.x
RHS headerless binary support 3.x

See:
  * http://intantech.com/files/Intan_RHD2000_data_file_formats.pdf
  * http://intantech.com/files/Intan_RHS2000_data_file_formats.pdf

Author: Samuel Garcia (Initial), Zach McKenzie & Heberto Mayorquin (Updates)

"""

from pathlib import Path
from packaging.version import Version
import warnings

import numpy as np

from neo.core import NeoReadWriteError

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)


class IntanRawIO(BaseRawIO):
    """
    Class for reading rhd and rhs Intan data

    Parameters
    ----------
    filename: str, default: ''
         name of the 'rhd' or 'rhs' data/header file
    ignore_integrity_checks: bool, default: False
        If True, data that violates integrity assumptions will be loaded. At the moment the only integrity
        check we perform is that timestamps are continuous. Setting this to True will ignore this check and set
        the attribute `discontinuous_timestamps` to True if the timestamps are not continous. This attribute can be checked
        after parsing the header to see if the timestamps are continuous or not.
    Notes
    -----
    * The Intan reader can handle two file formats 'rhd' and 'rhs'. It will automatically
    check for the file extension and will gather the header information based on the
    extension. Additionally it functions with RHS v 1.0 and v 3.x and RHD 1.0, 1.1, 1.2, 1.3, 2.0,
    3.x files.

    * The Intan reader can also handle the headerless binary formats 'one-file-per-signal' and
    'one-file-per-channel' which have a header file called 'info.rhd' or 'info.rhs' and a series
    of binary files with the '.dat' suffix

    * The reader can handle three file formats 'header-attached', 'one-file-per-signal' and
    'one-file-per-channel'.

    * Intan files contain amplifier channels labeled 'A', 'B' 'C' or 'D'
    depending on the port in which they were recorded along with the following
    additional streams.
    0: 'RHD2000' amplifier channel
    1: 'RHD2000 auxiliary input channel',
    2: 'RHD2000 supply voltage channel',
    3: 'USB board ADC input channel',
    4: 'USB board digital input channel',
    5: 'USB board digital output channel'

    And for RHS:

    0: 'RHS2000 amplfier channel'
    3: 'USB board ADC input channel',
    4: 'USB board ADC output channel',
    5: 'USB board digital input channel',
    6: 'USB board digital output channel',
    10: 'DC Amplifier channel',
    11: 'Stim channel',

    * For the "header-attached" and "one-file-per-signal" formats, the structure of the digital input and output channels is
    one long vector, which must be post-processed to extract individual digital channel information.
    See the intantech website for more information on performing this post-processing.

    Examples
    --------
    >>> import neo.rawio
    >>> # for a header-attached file
    >>> reader = neo.rawio.IntanRawIO(filename='data.rhd')
    >>> # for the other formats we point to the info.rhd
    >>> reader = neo.rawioIntanRawIO(filename='info.rhd')
    >>> reader.parse_header()
    >>> raw_chunk = reader.get_analogsignal_chunk(block_index=0,
                                                  seg_index=0
                                                  stream_index=0)
    >>> float_chunk = reader.rescale_signal_raw_to_float(raw_chunk, stream_index=0)

    """

    extensions = ["rhd", "rhs", "dat"]
    rawmode = "one-file"

    def __init__(self, filename="", ignore_integrity_checks=False):

        BaseRawIO.__init__(self)
        self.filename = Path(filename)
        self.ignore_integrity_checks = ignore_integrity_checks
        self.discontinuous_timestamps = False

    def _source_name(self):
        return self.filename

    def _parse_header(self):

        self.filename = Path(self.filename)

        # Input checks
        if not self.filename.is_file():
            raise FileNotFoundError(f"{self.filename} does not exist")

        if not self.filename.suffix in [".rhd", ".rhs"]:
            raise ValueError(f"{self.filename} is not a valid Intan file. Expected .rhd or .rhs extension")

        # see comment below for RHD which explains the division between file types
        if self.filename.suffix == ".rhs":
            if self.filename.name == "info.rhs":
                if any((self.filename.parent / file).exists() for file in one_file_per_signal_filenames_rhs):
                    self.file_format = "one-file-per-signal"
                    raw_file_paths_dict = create_one_file_per_signal_dict_rhs(dirname=self.filename.parent)
                else:
                    self.file_format = "one-file-per-channel"
                    raw_file_paths_dict = create_one_file_per_channel_dict_rhs(dirname=self.filename.parent)
            else:
                self.file_format = "header-attached"

            (
                self._global_info,
                self._ordered_channel_info,
                memmap_data_dtype,
                header_size,
                self._block_size,
                channel_number_dict,
            ) = read_rhs(self.filename, self.file_format)

        # 3 possibilities for rhd files, one combines the header and the data in the same file with suffix `rhd` while
        # the other two separates the data from the header which is always called `info.rhd`
        # attached to the actual binary file with data
        elif self.filename.suffix == ".rhd":
            if self.filename.name == "info.rhd":
                # first we have one-file-per-signal which is where one neo stream/file is saved as .dat files
                if any((self.filename.parent / file).exists() for file in one_file_per_signal_filenames_rhd):
                    self.file_format = "one-file-per-signal"
                    raw_file_paths_dict = create_one_file_per_signal_dict_rhd(dirname=self.filename.parent)
                # then there is one-file-per-channel where each channel in a neo stream is in its own .dat file
                else:
                    self.file_format = "one-file-per-channel"
                    raw_file_paths_dict = create_one_file_per_channel_dict_rhd(dirname=self.filename.parent)
            # finally the format with the header-attached to the binary file as one giant file
            else:
                self.file_format = "header-attached"

            (
                self._global_info,
                self._ordered_channel_info,
                memmap_data_dtype,
                header_size,
                self._block_size,
                channel_number_dict,
            ) = read_rhd(self.filename, self.file_format)

        # memmap the raw data for each format type
        # if header-attached there is one giant memory-map
        if self.file_format == "header-attached":
            self._raw_data = np.memmap(self.filename, dtype=memmap_data_dtype, mode="r", offset=header_size)

        # for 'one-file-per-signal' we have one memory map / neo stream
        elif self.file_format == "one-file-per-signal":
            self._raw_data = {}
            for stream_index, (stream_index_key, stream_datatype) in enumerate(memmap_data_dtype.items()):
                num_channels = channel_number_dict[stream_index_key]
                file_path = raw_file_paths_dict[stream_index_key]
                size_in_bytes = file_path.stat().st_size
                dtype_size = np.dtype(stream_datatype).itemsize
                n_samples = size_in_bytes // (dtype_size * num_channels)
                signal_stream_memmap = np.memmap(
                    file_path, dtype=stream_datatype, mode="r", shape=(num_channels, n_samples)
                ).T
                self._raw_data[stream_index] = signal_stream_memmap

        # for one-file-per-channel we have one memory map / channel stored as a list / neo stream
        elif self.file_format == "one-file-per-channel":
            self._raw_data = {}
            for stream_index, (stream_index_key, stream_datatype) in enumerate(memmap_data_dtype.items()):
                self._raw_data[stream_index] = []
                num_channels = channel_number_dict[stream_index_key]
                for channel_index in range(num_channels):
                    file_path = raw_file_paths_dict[stream_index_key][channel_index]
                    channel_memmap = np.memmap(file_path, dtype=stream_datatype, mode="r")
                    self._raw_data[stream_index].append(channel_memmap)

        # Data Integrity checks
        # strictness of check is controlled by ignore_integrity_checks
        # which is set at __init__
        self._assert_timestamp_continuity()

        # signals
        signal_channels = []
        self.native_channel_order = {}
        for chan_info in self._ordered_channel_info:
            name = chan_info["custom_channel_name"]
            channel_id = chan_info["native_channel_name"]
            sig_dtype = chan_info["dtype"]
            stream_id = str(chan_info["signal_type"])
            signal_channels.append(
                (
                    name,
                    channel_id,
                    chan_info["sampling_rate"],
                    sig_dtype,
                    chan_info["units"],
                    chan_info["gain"],
                    chan_info["offset"],
                    stream_id,
                )
            )
            self.native_channel_order[channel_id] = chan_info["native_order"]
        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        stream_ids = np.unique(signal_channels["stream_id"])
        signal_streams = np.zeros(stream_ids.size, dtype=_signal_stream_dtype)

        # we need to sort the data because the string of stream_index 10 is mis-sorted.
        stream_ids_sorted = sorted([int(stream_id) for stream_id in stream_ids])
        signal_streams["id"] = [str(stream_id) for stream_id in stream_ids_sorted]

        for stream_index, stream_id in enumerate(stream_ids_sorted):
            if self.filename.suffix == ".rhd":
                name = stream_id_to_stream_name_rhd.get(int(stream_id), "")
            else:
                name = stream_id_to_stream_name_rhs.get(int(stream_id), "")

            signal_streams["name"][stream_index] = name

        self._max_sampling_rate = np.max(signal_channels["sampling_rate"])

        # if header is attached we need to incorporate our block size to get signal length
        if self.file_format == "header-attached":
            self._max_sigs_length = self._raw_data.size * self._block_size
        # for one-file-per-signal we just take the size which will give n_samples for each
        # signal stream and then we just take the longest one
        elif self.file_format == "one-file-per-signal":
            self._max_sigs_length = max([raw_data.size for raw_data in self._raw_data.values()])
        # for one-file-per-channel we do the same as for one-file-per-signal, but since they
        # are in a list we just take the first channel in each list of channels
        else:
            self._max_sigs_length = max([raw_data[0].size for raw_data in self._raw_data.values()])

        # No events
        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # fill into header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # Extract annotations from the format
        self._generate_minimal_annotations()
        bl_annotations = self.raw_annotations["blocks"][0]
        seg_annotations = bl_annotations["segments"][0]

        for signal_annotation in seg_annotations["signals"]:
            # Add global annotations
            signal_annotation["intan_version"] = (
                f"{self._global_info['major_version']}." f"{self._global_info['minor_version']}"
            )
            global_keys_to_skip = [
                "major_version",
                "minor_version",
                "sampling_rate",
                "magic_number",
                "reference_channel",
            ]
            global_keys_to_annotate = set(self._global_info.keys()) - set(global_keys_to_skip)
            for key in global_keys_to_annotate:
                signal_annotation[key] = self._global_info[key]

            reference_channel = self._global_info.get("reference_channel", None)
            # Following the pdf specification
            reference_channel = "hardware" if reference_channel == "n/a" else reference_channel

            # Add channel annotations
            array_annotations = signal_annotation["__array_annotations__"]
            channel_ids = array_annotations["channel_ids"]

            # TODO refactor ordered channel dict to make this easier
            # Use this to find which elements of the ordered channels correspond to the current signal
            signal_type = int(signal_annotation["stream_id"])
            channel_info = next((info for info in self._ordered_channel_info if info["signal_type"] == signal_type))
            channel_keys_to_skip = [
                "signal_type",
                "custom_channel_name",
                "native_channel_name",
                "gain",
                "offset",
                "channel_enabled",
                "dtype",
                "units",
                "sampling_rate",
            ]

            channel_keys_to_annotate = set(channel_info.keys()) - set(channel_keys_to_skip)
            properties_dict = {key: [] for key in channel_keys_to_annotate}
            for channel_id in channel_ids:
                matching_info = next(
                    info for info in self._ordered_channel_info if info["native_channel_name"] == channel_id
                )
                for key in channel_keys_to_annotate:
                    properties_dict[key].append(matching_info[key])

            for key in channel_keys_to_annotate:
                array_annotations[key] = properties_dict[key]

    def _segment_t_start(self, block_index, seg_index):
        return 0.0

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self._max_sigs_length / self._max_sampling_rate
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):

        if self.file_format == "header-attached":
            stream_id = self.header["signal_streams"][stream_index]["id"]
            mask = self.header["signal_channels"]["stream_id"] == stream_id
            signal_channels = self.header["signal_channels"][mask]
            channel_ids = signal_channels["id"]

            stream_name = self.header["signal_streams"][stream_index]["name"][:]
            stream_is_digital = stream_name in digital_stream_names
            field_name = stream_name if stream_is_digital else channel_ids[0]

            size = self._raw_data[field_name].size

        # one-file-per-signal is (n_samples, n_channels)
        elif self.file_format == "one-file-per-signal":
            size = self._raw_data[stream_index].shape[0]
        # one-file-per-channel is (n_samples) so pull from list

        else:
            size = self._raw_data[stream_index][0].size

        return size

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return 0.0

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):

        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self._get_signal_size(block_index, seg_index, stream_index)

        if channel_indexes is None:
            channel_indexes = slice(None)

        if self.file_format == "header-attached":
            sigs_chunk = self._get_analogsignal_chunk_header_attached(
                i_start,
                i_stop,
                stream_index,
                channel_indexes,
            )
        elif self.file_format == "one-file-per-signal":
            sigs_chunk = self._get_analogsignal_chunk_one_file_per_signal(
                i_start,
                i_stop,
                stream_index,
                channel_indexes,
            )
        else:
            sigs_chunk = self._get_analogsignal_chunk_one_file_per_channel(
                i_start,
                i_stop,
                stream_index,
                channel_indexes,
            )

        return sigs_chunk

    def _get_analogsignal_chunk_header_attached(self, i_start, i_stop, stream_index, channel_indexes):

        stream_id = self.header["signal_streams"][stream_index]["id"]
        mask = self.header["signal_channels"]["stream_id"] == stream_id
        signal_channels = self.header["signal_channels"][mask]
        channel_ids = signal_channels["id"][channel_indexes]

        stream_name = self.header["signal_streams"][stream_index]["name"][:]
        stream_is_digital = stream_name in digital_stream_names

        field_name = stream_name if stream_is_digital else channel_ids[0]

        shape = self._raw_data[field_name].shape
        dtype = self._raw_data[field_name].dtype

        # This is False for Temperature and timestamps
        multiple_samples_per_block = len(shape) == 2

        # In the header attached case the data for each channel comes interleaved in blocks
        # To avoid unecessary memory access we can calculate the blocks we need to access beforehand:
        if multiple_samples_per_block:
            block_size = shape[1]
            block_start = i_start // block_size
            block_stop = i_stop // block_size + 1
            sl0 = i_start % block_size
            sl1 = sl0 + (i_stop - i_start)

        # For all streams raw_data is a structured memmap with a field for each channel_id
        if not stream_is_digital:
            sigs_chunk = np.zeros((i_stop - i_start, len(channel_ids)), dtype=dtype)

            for chunk_index, channel_id in enumerate(channel_ids):
                data_chan = self._raw_data[channel_id]
                if multiple_samples_per_block:
                    sigs_chunk[:, chunk_index] = data_chan[block_start:block_stop].flatten()[sl0:sl1]
                else:
                    sigs_chunk[:, chunk_index] = data_chan[i_start:i_stop]
        else:  # For digital data the channels come interleaved in a single field so we need to demultiplex
            digital_raw_data = self._raw_data[field_name].flatten()
            sigs_chunk = self._demultiplex_digital_data(digital_raw_data, channel_ids, i_start, i_stop)
        return sigs_chunk

    def _get_analogsignal_chunk_one_file_per_channel(self, i_start, i_stop, stream_index, channel_indexes):

        signal_data_memmap_list = self._raw_data[stream_index]
        channel_indexes_are_slice = isinstance(channel_indexes, slice)
        if channel_indexes_are_slice:
            num_channels = len(signal_data_memmap_list)
            start = channel_indexes.start or 0
            stop = channel_indexes.stop or num_channels
            step = channel_indexes.step or 1
            channel_indexes = range(start, stop, step)

        # We get the dtype from the first channel
        first_channel_index = channel_indexes[0]
        dtype = signal_data_memmap_list[first_channel_index].dtype
        sigs_chunk = np.zeros((i_stop - i_start, len(channel_indexes)), dtype=dtype)

        for chunk_index, channel_index in enumerate(channel_indexes):
            channel_memmap = signal_data_memmap_list[channel_index]
            sigs_chunk[:, chunk_index] = channel_memmap[i_start:i_stop]

        return sigs_chunk

    def _get_analogsignal_chunk_one_file_per_signal(self, i_start, i_stop, stream_index, channel_indexes):

        stream_name = self.header["signal_streams"][stream_index]["name"][:]
        raw_data = self._raw_data[stream_index]

        stream_is_digital = stream_name in digital_stream_names
        if stream_is_digital:
            stream_id = self.header["signal_streams"][stream_index]["id"]
            mask = self.header["signal_channels"]["stream_id"] == stream_id
            signal_channels = self.header["signal_channels"][mask]
            channel_ids = signal_channels["id"][channel_indexes]

            output = self._demultiplex_digital_data(raw_data, channel_ids, i_start, i_stop)

        else:
            output = raw_data[i_start:i_stop, channel_indexes]

        return output

    def _demultiplex_digital_data(self, raw_digital_data, channel_ids, i_start, i_stop):

        dtype = np.uint16  # We fix this to match the memmap dtype
        output = np.zeros((i_stop - i_start, len(channel_ids)), dtype=dtype)

        for channel_index, channel_id in enumerate(channel_ids):
            native_order = self.native_channel_order[channel_id]
            mask = 1 << native_order
            demultiplex_data = np.bitwise_and(raw_digital_data, mask) > 0
            output[:, channel_index] = demultiplex_data[i_start:i_stop].flatten()

        return output

    def _assert_timestamp_continuity(self):
        """
        Asserts the continuity of timestamps in the data.

        This method verifies that the timestamps in the raw data are sequential,
        indicating a continuous recording. If discontinuities are found, a flag
        is set to indicate potential data integrity issues, and an error is raised
        unless `ignore_integrity_checks` is True.

        Raises
        ------
        NeoReadWriteError
            If timestamps are not continuous and `ignore_integrity_checks` is False.
            The error message includes a table detailing the discontinuities found.

        Notes
        -----
        The method extracts timestamps from the raw data based on the file format:

        * **header-attached:** Timestamps are extracted from a 'timestamp' field in the raw data.
        * **one-file-per-signal:** Timestamps are taken from the last stream.
        * **one-file-per-channel:** Timestamps are retrieved from the first channel of the last stream.
        """
        # check timestamp continuity
        if self.file_format == "header-attached":
            timestamp = self._raw_data["timestamp"].flatten()

        # timestamps are always last stream for headerless binary files
        elif self.file_format == "one-file-per-signal":
            time_stream_index = max(self._raw_data.keys())
            timestamp = self._raw_data[time_stream_index]
        elif self.file_format == "one-file-per-channel":
            time_stream_index = max(self._raw_data.keys())
            timestamp = self._raw_data[time_stream_index][0]

        discontinuous_timestamps = np.diff(timestamp) != 1
        timestamps_are_not_contiguous = np.any(discontinuous_timestamps)
        if timestamps_are_not_contiguous:
            # Mark a flag that can be checked after parsing the header to see if the timestamps are continuous or not
            self.discontinuous_timestamps = True
            if not self.ignore_integrity_checks:
                error_msg = (
                    "\nTimestamps are not continuous, likely due to a corrupted file or inappropriate file merge.\n"
                    "To open the file anyway, initialize the reader with `ignore_integrity_checks=True`.\n\n"
                    "Discontinuities Found:\n"
                    "+-----------------+-----------------+-----------------+-----------------------+\n"
                    "| Discontinuity   | Previous        | Next            | Time Difference       |\n"
                    "| Index           | (Frames)        | (Frames)        | (Seconds)             |\n"
                    "+-----------------+-----------------+-----------------+-----------------------+\n"
                )

                amplifier_sampling_rate = self._global_info["sampling_rate"]
                for discontinuity_index in np.where(discontinuous_timestamps)[0]:
                    prev_ts = timestamp[discontinuity_index]
                    next_ts = timestamp[discontinuity_index + 1]
                    time_diff = (next_ts - prev_ts) / amplifier_sampling_rate

                    error_msg += (
                        f"| {discontinuity_index + 1:>15,} | {prev_ts:>15,} | {next_ts:>15,} | {time_diff:>21.6f} |\n"
                    )

                error_msg += "+-----------------+-----------------+-----------------+-----------------------+\n"

                raise NeoReadWriteError(error_msg)


###################################
# Header reading helper functions


def read_qstring(f):
    """
    Reads the optional notes included in the Intan RHX software

    Parameters
    ----------
    f: BinaryIO
        The file object

    Returns
    -------
    txt: str
        The string
    """
    length = np.fromfile(f, dtype="uint32", count=1)[0]
    if length == 0xFFFFFFFF or length == 0:
        return ""
    txt = f.read(length).decode("utf-16")
    return txt


def read_variable_header(f, header: list):
    """
    Reads the information from the binary file for the header info dict

    Parameters
    ----------
    f: BinaryIO
        The file object
    header: list[tuple]
        The list of header sections along with their associated dtype

    Returns
    -------
    info: dict
        The dictionary containing the information contained in the header
    """
    info = {}
    for field_name, field_type in header:
        if field_type == "QString":
            field_value = read_qstring(f)
        else:
            field_value = np.fromfile(f, dtype=field_type, count=1)[0]
        info[field_name] = field_value
    return info


###############
# RHS ZONE

rhs_global_header = [
    ("magic_number", "uint32"),  # 0xD69127AC
    ("major_version", "int16"),
    ("minor_version", "int16"),
    ("sampling_rate", "float32"),
    ("dsp_enabled", "int16"),
    ("actual_dsp_cutoff_frequency", "float32"),
    ("actual_lower_bandwidth", "float32"),
    ("actual_lower_settle_bandwidth", "float32"),
    ("actual_upper_bandwidth", "float32"),
    ("desired_dsp_cutoff_frequency", "float32"),
    ("desired_lower_bandwidth", "float32"),
    ("desired_lower_settle_bandwidth", "float32"),
    ("desired_upper_bandwidth", "float32"),
    ("notch_filter_mode", "int16"),
    ("desired_impedance_test_frequency", "float32"),
    ("actual_impedance_test_frequency", "float32"),
    ("amp_settle_mode", "int16"),
    ("charge_recovery_mode", "int16"),
    ("stim_step_size", "float32"),
    ("recovery_current_limit", "float32"),
    ("recovery_target_voltage", "float32"),
    ("note1", "QString"),
    ("note2", "QString"),
    ("note3", "QString"),
    ("dc_amplifier_data_saved", "int16"),
    ("board_mode", "int16"),
    ("ref_channel_name", "QString"),
    ("nb_signal_group", "int16"),
]

rhs_signal_group_header = [
    ("signal_group_name", "QString"),
    ("signal_group_prefix", "QString"),
    ("signal_group_enabled", "int16"),
    ("channel_num", "int16"),
    ("amplified_channel_num", "int16"),
]

rhs_signal_channel_header = [
    ("native_channel_name", "QString"),
    ("custom_channel_name", "QString"),
    ("native_order", "int16"),
    ("custom_order", "int16"),
    ("signal_type", "int16"),
    ("channel_enabled", "int16"),
    ("chip_channel_num", "int16"),
    ("command_stream", "int16"),
    ("board_stream_num", "int16"),
    ("spike_scope_trigger_mode", "int16"),
    ("spike_scope_voltage_thresh", "int16"),
    ("spike_scope_digital_trigger_channel", "int16"),
    ("spike_scope_digital_edge_polarity", "int16"),
    ("electrode_impedance_magnitude", "float32"),
    ("electrode_impedance_phase", "float32"),
]

stream_id_to_stream_name_rhs = {
    0: "RHS2000 amplifier channel",
    3: "USB board ADC input channel",
    4: "USB board ADC output channel",
    5: "USB board digital input channel",
    6: "USB board digital output channel",
    10: "DC Amplifier channel",
    11: "Stim channel",
}


def read_rhs(filename, file_format: str):
    BLOCK_SIZE = 128  # sample per block

    with open(filename, mode="rb") as f:
        global_info = read_variable_header(f, rhs_global_header)

        # We use signal_type in the header as stream_id in neo with the following
        # complications: for rhs files in the header-attaached stream_id 0 also
        # contains information for stream_id 10 and stream_id 11 so we need to
        # break these up. See notes throughout code base; for timestamps we always
        # force them to be the last stream_id.
        stream_id_to_channel_info_list = {k: [] for k in [0, 3, 4, 5, 6]}
        if not file_format == "header-attached":
            # memmap_data_dtype for rhs is complicated. There is not 1, 2 (supply and aux),
            # but there are dc-amp (10) and stim (11). we make timestamps (15)
            memmap_data_dtype = {k: [] for k in [0, 3, 4, 5, 6, 10, 11, 15]}
        for g in range(global_info["nb_signal_group"]):
            group_info = read_variable_header(f, rhs_signal_group_header)

            if bool(group_info["signal_group_enabled"]):
                for c in range(group_info["channel_num"]):
                    chan_info = read_variable_header(f, rhs_signal_channel_header)
                    if chan_info["signal_type"] in (1, 2):
                        error_msg = (
                            "signal_type of 1 or 2 does not exist for rhs files. If you have an rhs file "
                            "with these formats open an issue on the python-neo github page"
                        )
                        raise NeoReadWriteError(error_msg)
                    if bool(chan_info["channel_enabled"]):
                        stream_id_to_channel_info_list[chan_info["signal_type"]].append(chan_info)

        # useful dictionary for knowing the number of channels for non-header attached formats
        channel_number_dict = {i: len(stream_id_to_channel_info_list[i]) for i in [0, 3, 4, 5, 6]}

        header_size = f.tell()

    sr = global_info["sampling_rate"]

    # construct dtype by re-ordering channels by types
    ordered_channel_info = []
    if file_format == "header-attached":
        memmap_data_dtype = [("timestamp", "int32", BLOCK_SIZE)]
    else:
        memmap_data_dtype[15] = "int32"
        channel_number_dict[15] = 1

    # 0: RHS2000 amplifier channel.
    for chan_info in stream_id_to_channel_info_list[0]:
        chan_info["sampling_rate"] = sr
        chan_info["units"] = "uV"
        chan_info["gain"] = 0.195
        if file_format == "header-attached":
            chan_info["offset"] = -32768 * 0.195
        else:
            chan_info["offset"] = 0.0
        if file_format == "header-attached":
            chan_info["dtype"] = "uint16"
        else:
            chan_info["dtype"] = "int16"
        ordered_channel_info.append(chan_info)
        if file_format == "header-attached":
            name = chan_info["native_channel_name"]
            memmap_data_dtype += [(name, "uint16", BLOCK_SIZE)]
        else:
            memmap_data_dtype[0] = "int16"

    if bool(global_info["dc_amplifier_data_saved"]):
        # if we have dc amp we need to grab the correct number of channels
        # the channel number is the same as the count for amplifier data
        channel_number_dict[10] = channel_number_dict[0]
        for chan_info in stream_id_to_channel_info_list[0]:
            chan_info_dc = dict(chan_info)
            name = chan_info["native_channel_name"]
            chan_info_dc["native_channel_name"] = name + "_DC"
            chan_info_dc["sampling_rate"] = sr
            chan_info_dc["units"] = "mV"
            chan_info_dc["gain"] = 19.23
            chan_info_dc["offset"] = -512 * 19.23
            chan_info_dc["signal_type"] = 10  # put it in another group
            chan_info_dc["dtype"] = "uint16"
            ordered_channel_info.append(chan_info_dc)
            if file_format == "header-attached":
                memmap_data_dtype += [(name + "_DC", "uint16", BLOCK_SIZE)]
            else:
                memmap_data_dtype[10] = "uint16"

    # I can't seem to get stim files to generate for one-file-per-channel
    # so let's skip for now and can be given on request

    if file_format != "one-file-per-channel":
        channel_number_dict[11] = channel_number_dict[0]  # should be one stim / amplifier channel
        for chan_info in stream_id_to_channel_info_list[0]:
            chan_info_stim = dict(chan_info)
            name = chan_info["native_channel_name"]
            chan_info_stim["native_channel_name"] = name + "_STIM"
            chan_info_stim["sampling_rate"] = sr
            # stim channel are complicated because they are coded
            # with bits, they do not fit the gain/offset rawio strategy
            chan_info_stim["units"] = ""
            chan_info_stim["gain"] = 1.0
            chan_info_stim["offset"] = 0.0
            chan_info_stim["signal_type"] = 11  # put it in another group
            chan_info_stim["dtype"] = "uint16"
            ordered_channel_info.append(chan_info_stim)
            if file_format == "header-attached":
                memmap_data_dtype += [(name + "_STIM", "uint16", BLOCK_SIZE)]
            else:
                memmap_data_dtype[11] = "uint16"
    else:
        warnings.warn("Stim not implemented for `one-file-per-channel` due to lack of test files")

    # No supply or aux for rhs files (ie no stream_id 1 and 2)
    # We have an error above that requests test files to help if the spec is changed
    # in the future.

    # 3: Analog input channel.
    # 4: Analog output channel.
    for stream_id in [3, 4]:
        for chan_info in stream_id_to_channel_info_list[stream_id]:
            chan_info["sampling_rate"] = sr
            chan_info["units"] = "V"
            chan_info["gain"] = 0.0003125
            chan_info["offset"] = -32768 * 0.0003125
            chan_info["dtype"] = "uint16"
            ordered_channel_info.append(chan_info)
            if file_format == "header-attached":
                name = chan_info["native_channel_name"]
                memmap_data_dtype += [(name, "uint16", BLOCK_SIZE)]
            else:
                memmap_data_dtype[stream_id] = "uint16"

    # 5: Digital input channel.
    # 6: Digital output channel.
    for stream_id in [5, 6]:
        for chan_info in stream_id_to_channel_info_list[stream_id]:
            chan_info["sampling_rate"] = sr
            # arbitrary units are used to indicate that Intan does not
            # store raw voltages but only the boolean TTL state
            chan_info["units"] = "a.u."
            chan_info["gain"] = 1.0
            chan_info["offset"] = 0.0
            chan_info["dtype"] = "uint16"
            ordered_channel_info.append(chan_info)

        # Note that all the channels are packed in one buffer, so the data type only needs to be added once
        if len(stream_id_to_channel_info_list[stream_id]) > 0:
            if file_format == "header-attached":
                name = stream_id_to_stream_name_rhs[stream_id]
                memmap_data_dtype += [(name, "uint16", BLOCK_SIZE)]
            elif file_format == "one-file-per-channel":
                memmap_data_dtype[stream_id] = "uint16"
            elif file_format == "one-file-per-signal":
                memmap_data_dtype[stream_id] = "uint16"

    # per discussion with Intan developers before version 3 of their software the 'notch_filter_mode'
    # was a request for postprocessing to be done in one of their scripts. From version 3+ the notch
    # filter is now applied to the data in realtime and only the post notched amplifier data is
    # saved.
    if global_info["notch_filter_mode"] == 2 and global_info["major_version"] >= 3:
        global_info["notch_filter"] = "60Hz"
    elif global_info["notch_filter_mode"] == 1 and global_info["major_version"] >= 3:
        global_info["notch_filter"] = "50Hz"
    else:
        global_info["notch_filter"] = False

    if not file_format == "header-attached":
        # filter out dtypes without any values
        memmap_data_dtype = {k: v for (k, v) in memmap_data_dtype.items() if len(v) > 0}
        channel_number_dict = {k: v for (k, v) in channel_number_dict.items() if v > 0}

    return global_info, ordered_channel_info, memmap_data_dtype, header_size, BLOCK_SIZE, channel_number_dict


###############
# RHD ZONE

rhd_global_header_base = [
    ("magic_number", "uint32"),  # 0xC6912702
    ("major_version", "int16"),
    ("minor_version", "int16"),
]

rhd_global_header_part1 = [
    ("sampling_rate", "float32"),
    ("dsp_enabled", "int16"),
    ("actual_dsp_cutoff_frequency", "float32"),
    ("actual_lower_bandwidth", "float32"),
    ("actual_upper_bandwidth", "float32"),
    ("desired_dsp_cutoff_frequency", "float32"),
    ("desired_lower_bandwidth", "float32"),
    ("desired_upper_bandwidth", "float32"),
    ("notch_filter_mode", "int16"),
    ("desired_impedance_test_frequency", "float32"),
    ("actual_impedance_test_frequency", "float32"),
    ("note1", "QString"),
    ("note2", "QString"),
    ("note3", "QString"),
]

rhd_global_header_v11 = [
    ("num_temp_sensor_channels", "int16"),
]

rhd_global_header_v13 = [
    ("eval_board_mode", "int16"),
]

rhd_global_header_v20 = [
    ("reference_channel", "QString"),
]

rhd_global_header_final = [
    ("nb_signal_group", "int16"),
]

rhd_signal_group_header = [
    ("signal_group_name", "QString"),
    ("signal_group_prefix", "QString"),
    ("signal_group_enabled", "int16"),
    ("channel_num", "int16"),
    ("amplified_channel_num", "int16"),
]

rhd_signal_channel_header = [
    ("native_channel_name", "QString"),
    ("custom_channel_name", "QString"),
    ("native_order", "int16"),
    ("custom_order", "int16"),
    ("signal_type", "int16"),
    ("channel_enabled", "int16"),
    ("chip_channel_num", "int16"),
    ("board_stream_num", "int16"),
    ("spike_scope_trigger_mode", "int16"),
    ("spike_scope_voltage_thresh", "int16"),
    ("spike_scope_digital_trigger_channel", "int16"),
    ("spike_scope_digital_edge_polarity", "int16"),
    ("electrode_impedance_magnitude", "float32"),
    ("electrode_impedance_phase", "float32"),
]

stream_id_to_stream_name_rhd = {
    0: "RHD2000 amplifier channel",
    1: "RHD2000 auxiliary input channel",
    2: "RHD2000 supply voltage channel",
    3: "USB board ADC input channel",
    4: "USB board digital input channel",
    5: "USB board digital output channel",
}


def read_rhd(filename, file_format: str):
    """Function for reading the rhd file header

    Parameters
    ----------
    filename: str | Path
        The filename of the *.rhd file
    file_format: 'header-attached' | 'one-file-per-signal' | 'one-file-per-channel'
        Whether the header is included with the rest of the data ('header-attached')
        Or as a standalone file ('one-file-per-signal' or 'one-file-per-channel')
    """
    with open(filename, mode="rb") as f:

        global_info = read_variable_header(f, rhd_global_header_base)

        version = Version(f"{global_info['major_version']}.{global_info['minor_version']}")

        # the header size depends on the version :-(
        header = list(rhd_global_header_part1)  # make a copy

        if version >= Version("1.1"):
            header = header + rhd_global_header_v11
        else:
            global_info["num_temp_sensor_channels"] = 0

        if version >= Version("1.3"):
            header = header + rhd_global_header_v13
        else:
            global_info["eval_board_mode"] = 0

        if version >= Version("2.0"):
            header = header + rhd_global_header_v20
        else:
            global_info["reference_channel"] = ""

        header = header + rhd_global_header_final

        global_info.update(read_variable_header(f, header))

        # read channel group and channel header
        stream_id_to_channel_info_list = {k: [] for k in [0, 1, 2, 3, 4, 5]}
        if not file_format == "header-attached":
            memmap_data_dtype = {k: [] for k in range(7)}  # 5 streams + 6 for timestamps for not header attached
        for g in range(global_info["nb_signal_group"]):
            group_info = read_variable_header(f, rhd_signal_group_header)

            if bool(group_info["signal_group_enabled"]):
                for c in range(group_info["channel_num"]):
                    chan_info = read_variable_header(f, rhd_signal_channel_header)
                    if bool(chan_info["channel_enabled"]):
                        stream_id_to_channel_info_list[chan_info["signal_type"]].append(chan_info)

            channel_number_dict = {i: len(stream_id_to_channel_info_list[i]) for i in range(6)}

        header_size = f.tell()

    sr = global_info["sampling_rate"]

    # construct the data block dtype and reorder channels
    if version >= Version("2.0"):
        BLOCK_SIZE = 128
    else:
        BLOCK_SIZE = 60  # 256 channels

    ordered_channel_info = []

    if version >= Version("1.2"):
        if file_format == "header-attached":
            memmap_data_dtype = [("timestamp", "int32", BLOCK_SIZE)]
        else:
            memmap_data_dtype[6] = "int32"
            channel_number_dict[6] = 1
    else:
        if file_format == "header-attached":
            memmap_data_dtype = [("timestamp", "uint32", BLOCK_SIZE)]
        else:
            memmap_data_dtype[6] = "uint32"
            channel_number_dict[6] = 1

    # 0: RHD2000 amplifier channel
    for chan_info in stream_id_to_channel_info_list[0]:
        chan_info["sampling_rate"] = sr
        chan_info["units"] = "uV"
        chan_info["gain"] = 0.195
        if file_format == "header-attached":
            chan_info["offset"] = -32768 * 0.195
            chan_info["dtype"] = "uint16"
        else:
            chan_info["offset"] = 0.0
            chan_info["dtype"] = "int16"
        ordered_channel_info.append(chan_info)

        if file_format == "header-attached":
            name = chan_info["native_channel_name"]
            memmap_data_dtype += [(name, "uint16", BLOCK_SIZE)]
        else:
            memmap_data_dtype[0] = "int16"

    # 1: RHD2000 auxiliary input channel
    for chan_info in stream_id_to_channel_info_list[1]:
        chan_info["sampling_rate"] = sr / 4.0
        chan_info["units"] = "V"
        chan_info["gain"] = 0.0000374
        chan_info["offset"] = 0.0
        chan_info["dtype"] = "uint16"
        ordered_channel_info.append(chan_info)
        if file_format == "header-attached":
            name = chan_info["native_channel_name"]
            memmap_data_dtype += [(name, "uint16", BLOCK_SIZE // 4)]
        else:
            memmap_data_dtype[1] = "uint16"

    # 2: RHD2000 supply voltage channel
    for chan_info in stream_id_to_channel_info_list[2]:
        chan_info["sampling_rate"] = sr / BLOCK_SIZE
        chan_info["units"] = "V"
        chan_info["gain"] = 0.0000748
        chan_info["offset"] = 0.0
        chan_info["dtype"] = "uint16"
        ordered_channel_info.append(chan_info)
        if file_format == "header-attached":
            name = chan_info["native_channel_name"]
            memmap_data_dtype += [(name, "uint16")]
        else:
            memmap_data_dtype[2] = "uint16"

    # temperature is not an official channel in the header
    for i in range(global_info["num_temp_sensor_channels"]):
        name = f"temperature_{i}"
        chan_info = {"native_channel_name": name, "signal_type": 20}
        chan_info["sampling_rate"] = sr / BLOCK_SIZE
        chan_info["units"] = "Celsius"
        chan_info["gain"] = 0.001
        chan_info["offset"] = 0.0
        chan_info["dtype"] = "int16"
        ordered_channel_info.append(chan_info)
        memmap_data_dtype += [(name, "int16")]

    # 3: USB board ADC input channel
    for chan_info in stream_id_to_channel_info_list[3]:
        chan_info["sampling_rate"] = sr
        chan_info["units"] = "V"
        if global_info["eval_board_mode"] == 0:
            chan_info["gain"] = 0.000050354
            chan_info["offset"] = 0.0
        elif global_info["eval_board_mode"] == 1:
            chan_info["gain"] = 0.00015259
            chan_info["offset"] = -32768 * 0.00015259
        elif global_info["eval_board_mode"] == 13:
            chan_info["gain"] = 0.0003125
            chan_info["offset"] = -32768 * 0.0003125
        chan_info["dtype"] = "uint16"
        ordered_channel_info.append(chan_info)
        if file_format == "header-attached":
            name = chan_info["native_channel_name"]
            memmap_data_dtype += [(name, "uint16", BLOCK_SIZE)]
        else:
            memmap_data_dtype[3] = "uint16"

    # 4: USB board digital input channel
    # 5: USB board digital output channel
    for stream_id in [4, 5]:
        for chan_info in stream_id_to_channel_info_list[stream_id]:
            chan_info["sampling_rate"] = sr
            # arbitrary units are used to indicate that Intan does not
            # store raw voltages but only the boolean TTL state
            chan_info["units"] = "a.u."
            chan_info["gain"] = 1.0
            chan_info["offset"] = 0.0
            chan_info["dtype"] = "uint16"
            ordered_channel_info.append(chan_info)

        # Note that all the channels are packed in one buffer, so the data type only needs to be added once
        if len(stream_id_to_channel_info_list[stream_id]) > 0:
            if file_format == "header-attached":
                name = stream_id_to_stream_name_rhd[stream_id]
                memmap_data_dtype += [(name, "uint16", BLOCK_SIZE)]
            elif file_format == "one-file-per-channel":
                memmap_data_dtype[stream_id] = "uint16"
            elif file_format == "one-file-per-signal":
                memmap_data_dtype[stream_id] = "uint16"

    # per discussion with Intan developers before version 3 of their software the 'notch_filter_mode'
    # was a request for postprocessing to be done in one of their scripts. From version 3+ the notch
    # filter is now applied to the data in realtime and only the post notched amplifier data is
    # saved.
    if global_info["notch_filter_mode"] == 2 and version >= Version("3.0"):
        global_info["notch_filter"] = "60Hz"
    elif global_info["notch_filter_mode"] == 1 and version >= Version("3.0"):
        global_info["notch_filter"] = "50Hz"
    else:
        global_info["notch_filter"] = False

    if not file_format == "header-attached":
        # filter out dtypes without any values
        memmap_data_dtype = {k: v for (k, v) in memmap_data_dtype.items() if len(v) > 0}
        channel_number_dict = {k: v for (k, v) in channel_number_dict.items() if v > 0}

    return global_info, ordered_channel_info, memmap_data_dtype, header_size, BLOCK_SIZE, channel_number_dict


##########################################################################
# RHX Zone for Binary Files (note this is for version 3+ of Intan software)
# This section provides all possible headerless binary files in both the rhs and rhd
# formats.

digital_stream_names = ["USB board digital input channel", "USB board digital output channel"]


# RHD Binary Files for One File Per Signal
one_file_per_signal_filenames_rhd = [
    "amplifier.dat",
    "auxiliary.dat",
    "supply.dat",
    "analogin.dat",
    "digitalin.dat",
    "digitalout.dat",
]


def create_one_file_per_signal_dict_rhd(dirname):
    """Function for One File Per Signal Type

    Parameters
    ----------
    dirname: pathlib.Path
        The folder to explore

    Returns
    -------
    raw_files_paths_dict: dict
        A dict of all the file paths
    """

    raw_file_paths_dict = {}
    for raw_index, raw_file in enumerate(one_file_per_signal_filenames_rhd):
        if Path(dirname / raw_file).is_file():
            raw_file_paths_dict[raw_index] = Path(dirname / raw_file)

    raw_file_paths_dict[6] = Path(dirname / "time.dat")

    return raw_file_paths_dict


# RHS Binary Files for One File Per Signal
one_file_per_signal_filenames_rhs = [
    "amplifier.dat",
    "auxiliary.dat",
    "supply.dat",
    "analogin.dat",
    "analogout.dat",
    "digitalin.dat",
    "digitalout.dat",
]


def create_one_file_per_signal_dict_rhs(dirname):
    """Function for One File Per Signal Type

    Parameters
    ----------
    dirname: pathlib.Path
        The folder to explore

    Returns
    -------
    raw_files_paths_dict: dict
        A dict of all the file paths
    """

    raw_file_paths_dict = {}
    for raw_index, raw_file in enumerate(one_file_per_signal_filenames_rhs):
        if Path(dirname / raw_file).is_file():
            raw_file_paths_dict[raw_index] = Path(dirname / raw_file)

    # we need time to be the last value
    raw_file_paths_dict[15] = Path(dirname / "time.dat")
    # 10 and 11 are hardcoded in the rhs_reader above so hardcoded here too
    if Path(dirname / "dcamplifier.dat").is_file():
        raw_file_paths_dict[10] = Path(dirname / "dcamplifier.dat")
    if Path(dirname / "stim.dat").is_file():
        raw_file_paths_dict[11] = Path(dirname / "stim.dat")

    return raw_file_paths_dict


# RHD Binary Files for One File Per Channel
possible_raw_file_prefixes_rhd = [
    "amp",
    "aux",
    "vdd",
    "board-ANALOG-IN",
    "board-DIGITAL-IN",
    "board-DIGITAL-OUT",
]


def create_one_file_per_channel_dict_rhd(dirname):
    """Utility function for One File Per Channel

    Parameters
    ----------
    dirname: pathlib.Path
        The folder to explore

    Returns
    -------
    raw_files_paths_dict: dict
        A dict of all the file paths
    """

    file_names = dirname.glob("**/*.dat")
    files = [file for file in file_names if file.is_file()]
    raw_file_paths_dict = {}
    for raw_index, prefix in enumerate(possible_raw_file_prefixes_rhd):
        raw_file_paths_dict[raw_index] = [file for file in files if prefix in file.name]

    raw_file_paths_dict[6] = [Path(dirname / "time.dat")]

    return raw_file_paths_dict


# RHS Binary Files for One File Per Channel
possible_raw_file_prefixes_rhs = [
    "amp",
    "aux",
    "vdd",
    "board-ANALOG-IN",
    "board-ANALOG-OUT",
    "board-DIGITAL-IN",
    "board-DIGITAL-OUT",
]


def create_one_file_per_channel_dict_rhs(
    dirname,
):
    """Utility function for One File Per Channel

    Parameters
    ----------
    dirname: pathlib.Path
        The folder to explore

    Returns
    -------
    raw_files_paths_dict: dict
        A dict of all the file paths
    """

    file_names = dirname.glob("**/*.dat")
    files = [file for file in file_names if file.is_file()]
    raw_file_paths_dict = {}
    for raw_index, prefix in enumerate(possible_raw_file_prefixes_rhs):
        raw_file_paths_dict[raw_index] = [file for file in files if prefix in file.name]

    # we need time to be the last value
    raw_file_paths_dict[15] = [Path(dirname / "time.dat")]
    # 10 and 11 are hardcoded in the rhs reader so hardcoded here
    raw_file_paths_dict[10] = [file for file in files if "dc-" in file.name]
    # we can find the files, but I can see how to read them out of header
    # so for now we don't expose the stim files in one-file-per-channel
    raw_file_paths_dict[11] = [file for file in files if "stim-" in file.name]

    return raw_file_paths_dict



"""
Tests of neo.io.phyio
"""

import unittest

from neo.io.phyio import PhyIO  # , HAVE_SCIPY
from neo.test.iotest.common_io_test import BaseTestIO
from neo.io.proxyobjects import AnalogSignalProxy, SpikeTrainProxy, EventProxy, EpochProxy
from neo import AnalogSignal, SpikeTrain

import quantities as pq
import numpy as np

import tempfile
from pathlib import Path


class TestPhyIO(BaseTestIO, unittest.TestCase):
    ioclass = PhyIO
    entities_to_download = ["phy"]
    entities_to_test = ["phy/phy_example_0"]

    def test_read_segment_lazy(self):
        dirname = self.get_local_path("phy/phy_example_0")
        r = PhyIO(dirname=dirname)
        seg = r.read_segment(lazy=True)
        for ana in seg.analogsignals:
            assert isinstance(ana, AnalogSignalProxy)
            ana = ana.load()
            assert isinstance(ana, AnalogSignal)
        for st in seg.spiketrains:
            assert isinstance(st, SpikeTrainProxy)
            st = st.load()
            assert isinstance(st, SpikeTrain)

        seg = r.read_segment(lazy=False)
        for anasig in seg.analogsignals:
            assert isinstance(ana, AnalogSignal)
            self.assertNotEqual(anasig.size, 0)
        for st in seg.spiketrains:
            assert isinstance(st, SpikeTrain)
            self.assertNotEqual(st.size, 0)

        # annotations
        assert seg.name == "Seg #0 Block #0"
        for anasig in seg.analogsignals:
            assert anasig.name is not None
        for st in seg.spiketrains:
            assert st.name is not None
        for ev in seg.events:
            assert ev.name is not None
        for ep in seg.epochs:
            assert ep.name is not None

    def test_read_block(self):
        dirname = self.get_local_path("phy/phy_example_0")
        r = PhyIO(dirname=dirname)
        bl = r.read_block(lazy=True)

    def test_read_segment_with_time_slice(self):
        dirname = self.get_local_path("phy/phy_example_0")
        r = PhyIO(dirname=dirname)
        seg = r.read_segment(time_slice=None)
        spikes_full = seg.spiketrains[0]

        t_start, t_stop = 260 * pq.ms, 1.854 * pq.s
        seg = r.read_segment(time_slice=(t_start, t_stop))
        spikes_slice = seg.spiketrains[0]

        assert spikes_full.size > spikes_slice.size
        assert np.all(spikes_slice >= t_start)
        assert np.all(spikes_slice <= t_stop)
        assert spikes_slice.t_start == t_start
        assert spikes_slice.t_stop == t_stop


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.brainwaredamio
"""

import os.path

import unittest

import numpy as np
import quantities as pq

from neo.core import AnalogSignal, Block, Group, Segment
from neo.io import BrainwareDamIO
from neo.test.iotest.common_io_test import BaseTestIO
from neo.test.tools import assert_same_sub_schema, assert_neo_object_is_compliant
from neo.test.iotest.tools import create_generic_reader


def proc_dam(filename):
    """Load an dam file that has already been processed by the official matlab
    file converter.  That matlab data is saved to an m-file, which is then
    converted to a numpy '.npz' file.  This numpy file is the file actually
    loaded.  This function converts it to a neo block and returns the block.
    This block can be compared to the block produced by BrainwareDamIO to
    make sure BrainwareDamIO is working properly

    block = proc_dam(filename)

    filename: The file name of the numpy file to load.  It should end with
    '*_dam_py?.npz'. This will be converted to a neo 'file_origin' property
    with the value '*.dam', so the filename to compare should fit that pattern.
    'py?' should be 'py2' for the python 2 version of the numpy file or 'py3'
    for the python 3 version of the numpy file.

    example: filename = 'file1_dam_py2.npz'
             dam file name = 'file1.dam'
    """
    with np.load(filename, allow_pickle=True) as damobj:
        damfile = list(damobj.items())[0][1].flatten()

    filename = os.path.basename(filename[:-12] + ".dam")

    signals = [res.flatten() for res in damfile["signal"]]
    stimIndexes = [int(res[0, 0].tolist()) for res in damfile["stimIndex"]]
    timestamps = [res[0, 0] for res in damfile["timestamp"]]

    block = Block(file_origin=filename)

    gr = Group(file_origin=filename)

    block.groups.append(gr)

    params = [res["params"][0, 0].flatten() for res in damfile["stim"]]
    values = [res["values"][0, 0].flatten() for res in damfile["stim"]]
    params = [[res1[0] for res1 in res] for res in params]
    values = [[res1 for res1 in res] for res in values]
    stims = [dict(zip(param, value)) for param, value in zip(params, values)]

    fulldam = zip(stimIndexes, timestamps, signals, stims)
    for stimIndex, timestamp, signal, stim in fulldam:
        sig = AnalogSignal(
            signal=signal * pq.mV, t_start=timestamp * pq.d, file_origin=filename, sampling_period=1.0 * pq.s
        )
        segment = Segment(file_origin=filename, index=stimIndex, **stim)
        segment.analogsignals = [sig]
        block.segments.append(segment)
        gr.analogsignals.append(sig)
        sig.group = gr

    block.check_relationships()

    return block


class BrainwareDamIOTestCase(BaseTestIO, unittest.TestCase):
    """
    Unit test testcase for neo.io.BrainwareDamIO
    """

    ioclass = BrainwareDamIO
    read_and_write_is_bijective = False

    entities_to_download = ["brainwaredam"]

    # These are the files it tries to read and test for compliance
    entities_to_test = [
        "brainwaredam/block_300ms_4rep_1clust_part_ch1.dam",
        "brainwaredam/interleaved_500ms_5rep_ch2.dam",
        "brainwaredam/long_170s_1rep_1clust_ch2.dam",
        "brainwaredam/multi_500ms_mulitrep_ch1.dam",
        "brainwaredam/random_500ms_12rep_noclust_part_ch2.dam",
        "brainwaredam/sequence_500ms_5rep_ch2.dam",
    ]

    # these are reference files to compare to
    files_to_compare = [
        "brainwaredam/block_300ms_4rep_1clust_part_ch1",
        "brainwaredam/interleaved_500ms_5rep_ch2",
        "",
        "brainwaredam/multi_500ms_mulitrep_ch1",
        "brainwaredam/random_500ms_12rep_noclust_part_ch2",
        "brainwaredam/sequence_500ms_5rep_ch2",
    ]

    # add the suffix
    for i, fname in enumerate(files_to_compare):
        if fname:
            files_to_compare[i] += "_dam_py3.npz"

    def test_reading_same(self):
        for ioobj, path in self.iter_io_objects(return_path=True):
            obj_reader_base = create_generic_reader(ioobj, target=False)
            obj_reader_single = create_generic_reader(ioobj)

            obj_base = obj_reader_base()
            obj_single = obj_reader_single()

            try:
                assert_same_sub_schema(obj_base, [obj_single])
            except BaseException as exc:
                exc.args += ("from " + os.path.basename(path),)
                raise

    def test_against_reference(self):
        for filename, refname in zip(self.files_to_test, self.files_to_compare):
            if not refname:
                continue
            obj = self.read_file(filename=filename)
            refobj = proc_dam(self.get_local_path(refname))

            try:
                assert_neo_object_is_compliant(obj)
                assert_neo_object_is_compliant(refobj)
                assert_same_sub_schema(obj, refobj)
            except BaseException as exc:
                exc.args += ("from " + filename,)
                raise


if __name__ == "__main__":
    unittest.main()



"""
Tests of the neo.core.group.Group class and related functions
"""

import unittest

import numpy as np
import quantities as pq
from numpy.testing import assert_array_equal

from neo.core.analogsignal import AnalogSignal
from neo.core.irregularlysampledsignal import IrregularlySampledSignal
from neo.core.spiketrain import SpikeTrain
from neo.core.segment import Segment
from neo.core.view import ChannelView
from neo.core.group import Group
from neo.core.block import Block
from neo.core.imagesequence import ImageSequence
from neo.core.regionofinterest import CircularRegionOfInterest


class TestGroup(unittest.TestCase):

    def setUp(self):
        test_data = np.random.rand(100, 8) * pq.mV
        channel_names = np.array(["a", "b", "c", "d", "e", "f", "g", "h"])
        test_image_data = np.random.rand(640).reshape(10, 8, 8)
        self.test_signal = AnalogSignal(
            test_data,
            sampling_period=0.1 * pq.ms,
            name="test signal",
            description="this is a test signal",
            array_annotations={"channel_names": channel_names},
            attUQoLtUaE=42,
        )
        self.test_view = ChannelView(
            self.test_signal,
            [1, 2, 5, 7],
            name="view of test signal",
            description="this is a view of a test signal",
            array_annotations={"something": np.array(["A", "B", "C", "D"])},
            sLaTfat="fish",
        )
        self.test_image_seq = ImageSequence(test_image_data, frame_duration=20 * pq.ms, spatial_scale=1 * pq.um)
        self.roi = CircularRegionOfInterest(self.test_image_seq, 0, 0, 3)
        self.test_spiketrains = [
            SpikeTrain(np.arange(100.0), units="ms", t_stop=200),
            SpikeTrain(np.arange(0.5, 100.5), units="ms", t_stop=200),
        ]
        self.test_segment = Segment()
        self.test_segment.analogsignals.append(self.test_signal)
        self.test_segment.spiketrains.extend(self.test_spiketrains)
        self.test_segment.imagesequences.append(self.test_image_seq)

    def test_create_group(self):
        objects = [self.test_view, self.test_signal, self.test_image_seq, self.roi]
        objects.extend(self.test_spiketrains)
        group = Group(objects)

        assert group.analogsignals[0] is self.test_signal
        assert group.spiketrains[0] is self.test_spiketrains[0]
        assert group.spiketrains[1] is self.test_spiketrains[1]
        assert group.channelviews[0] is self.test_view
        assert group.imagesequences[0] is self.test_image_seq
        assert group.regionsofinterest[0] is self.roi
        assert len(group.irregularlysampledsignals) == 0

    def test_create_empty_group(self):
        group = Group()

    def test_children(self):
        group = Group(self.test_spiketrains + [self.test_view] + [self.test_signal])

        # note: ordering is by class name for data children (AnalogSignal, SpikeTrain),
        #       then container children (Segment)
        assert group.children == (self.test_signal, *self.test_spiketrains, self.test_view)

    def test_with_allowed_types(self):
        objects = [self.test_signal] + self.test_spiketrains
        group = Group(objects, allowed_types=(AnalogSignal, SpikeTrain))
        assert group.analogsignals[0] is self.test_signal
        assert group.spiketrains[0] is self.test_spiketrains[0]
        assert group.spiketrains[1] is self.test_spiketrains[1]
        self.assertRaises(TypeError, group.add, self.test_view)

    def test_walk(self):
        # set up nested groups
        parent = Group(name="0")
        children = (Group(name="00"), Group(name="01"), Group(name="02"))
        parent.add(*children)
        grandchildren = (
            (Group(name="000"), Group(name="001")),
            [],
            (Group(name="020"), Group(name="021"), Group(name="022")),
        )
        for child, gchildren in zip(children, grandchildren):
            child.add(*gchildren)

        flattened = list(parent.walk())
        target = [parent, children[0], *grandchildren[0]]
        target.extend([children[1], children[2], *grandchildren[2]])
        self.assertEqual(flattened, target)

    def test_add_invalid_type_raises_Exception(self):
        group = Group()
        self.assertRaises(TypeError, group.add, Block())

        self.assertRaises(TypeError, Group, allowed_types=[Block])
